static final int PERIODIC_BOUNDARY_CONDITIONS = 0;
static final int CLAMPED_BOUNDARY_CONDITIONS  = 1;
static final int CONSTANT_BOUNDARY_CONDITIONS = 2;
static final int BOUNDARY_CONDITIONS = PERIODIC_BOUNDARY_CONDITIONS;

PImage sanityCheck() {
  float[][] synthesis = new float[SZ][SZ];
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) synthesis[i][j] = fiducialScaleLayers.get(NUM_SCALES-1)[i][j];
  for (int k=0; k<NUM_SCALES-1; k++) for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) synthesis[i][j] += fiducialScaleLayers.get(k)[i][j];
  return dataToImage(synthesis);
}

int periodic(int i, int n) { return ((i % n) + n) % n; }

boolean inRange(int i) {
  return ((i>=0) && (i<SZ));
}

boolean inRange2D( int i, int j) {
  return (inRange(i) && inRange(j));
}

// Disarrays an image
PImage captcha(PImage anImage, int theLevel, float theReach) {
  float[][] data = imageToData(anImage);
  float[][] disarrayed = new float[SZ][SZ];
  ArrayList<float[][]> perturbation = twoRandomFields(theLevel);
  float[][] xComponent = perturbation.get(0);
  float[][] yComponent = perturbation.get(1);
  switch (BOUNDARY_CONDITIONS) {
    case PERIODIC_BOUNDARY_CONDITIONS:
      for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) 
          disarrayed[i][j] = data[periodic((int) (i + theReach*xComponent[i][j]),SZ)][periodic((int) (j + theReach*yComponent[i][j]),SZ)];
      break;
    case CLAMPED_BOUNDARY_CONDITIONS:
      for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) 
          disarrayed[i][j] = data[(int) constrain(i + theReach*xComponent[i][j],0,SZ-1)][(int) constrain(j + theReach*yComponent[i][j],0,SZ-1)];
      break;
    case CONSTANT_BOUNDARY_CONDITIONS:
      for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) {
        int ii = (int) (i + theReach*xComponent[i][j]);
        int jj = (int) (j + theReach*yComponent[i][j]);
        if (inRange2D(i,j)) disarrayed[i][j] = fiducialScaleLayers.get(theLevel)[ii][jj];
        else disarrayed[i][j] = 127.0;
      }
      break;
  }
  return dataToImage(disarrayed);
}

// Disarrays a single data plane
float[][] disarray(ArrayList<float[][]> theStackX, ArrayList<float[][]> theStackY, int theLevel, float theReach) {
  
  float[][] disarrayed = new float[SZ][SZ];
  
  switch (BOUNDARY_CONDITIONS) {
    
    case PERIODIC_BOUNDARY_CONDITIONS:
      for (int i=0; i<SZ; i++) 
        for (int j=0; j<SZ; j++) 
          disarrayed[i][j] = 
            fiducialScaleLayers.get(theLevel)[
                periodic((int) (i + theReach*theStackX.get(theLevel)[i][j]),SZ)]
              [
                periodic((int) (j + theReach*theStackY.get(theLevel)[i][j]),SZ)
              ];
      break;
      
    case CLAMPED_BOUNDARY_CONDITIONS:
      for (int i=0; i<SZ; i++) 
        for (int j=0; j<SZ; j++) 
          disarrayed[i][j] = 
            fiducialScaleLayers.get(theLevel)[
              (int) constrain(i + theReach*theStackX.get(theLevel)[i][j],0,SZ-1)
            ][
              (int) constrain(j + theReach*theStackY.get(theLevel)[i][j],0,SZ-1)
            ];
      break;
      
    case CONSTANT_BOUNDARY_CONDITIONS:
      for (int i=0; i<SZ; i++) 
        for (int j=0; j<SZ; j++) {
        int ii = (int) (i + theReach*theStackX.get(theLevel)[i][j]);
        int jj = (int) (j + theReach*theStackY.get(theLevel)[i][j]);
        if (inRange2D(i,j))
          disarrayed[i][j] = fiducialScaleLayers.get(theLevel)[ii][jj];
        else
          disarrayed[i][j] = 127.0;
      }
      break;
  }
  return disarrayed;
}

// Shows the effect of local disarray for a Lotze noise stack
PImage showLocalDisarrayLotze(int theLevel, float theReach) {
  
  ArrayList<float[][]> noiseStackX = new ArrayList<float[][]>();
  ArrayList<float[][]> noiseStackY = new ArrayList<float[][]>();
  for (int k=0; k<NUM_SCALES; k++) {
    ArrayList<float[][]> tmp = twoRandomFields(theLevel);
    noiseStackX.add(tmp.get(0));
    noiseStackY.add(tmp.get(1));
  }
  
  ArrayList<float[][]> tmp = new ArrayList<float[][]>();
  for (int k=0; k<NUM_SCALES; k++) tmp.add(disarray(noiseStackX,noiseStackY,k,theReach));
  float[][] synthesis = new float[SZ][SZ];
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) synthesis[i][j] = fiducialScaleLayers.get(NUM_SCALES-1)[i][j];
  for (int k=0; k<NUM_SCALES-1; k++) for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) synthesis[i][j] += tmp.get(k)[i][j];
  return dataToImage(synthesis);
}

PImage showLocalDisarrayHelmholtz(float theReach) {
  ArrayList<float[][]> tmp = new ArrayList<float[][]>();
  for (int k=0; k<NUM_SCALES; k++) tmp.add(disarray(randomGaussianScaledDataStackX,randomGaussianScaledDataStackY,k,theReach));
  float[][] synthesis = new float[SZ][SZ];
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) synthesis[i][j] = fiducialScaleLayers.get(NUM_SCALES-1)[i][j];
  for (int k=0; k<NUM_SCALES-1; k++) for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) synthesis[i][j] += tmp.get(k)[i][j];
  return dataToImage(synthesis);
}

PImage showLocalDisarrayCoherent(float theReach) {
  ArrayList<float[][]> tmp = new ArrayList<float[][]>();
  for (int k=0; k<NUM_SCALES; k++) tmp.add(disarray(coherentDataStackX,coherentDataStackY,k,theReach));
  float[][] synthesis = new float[SZ][SZ];
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) synthesis[i][j] = fiducialScaleLayers.get(NUM_SCALES-1)[i][j];
  for (int k=0; k<NUM_SCALES-1; k++) for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) synthesis[i][j] += tmp.get(k)[i][j];
  return dataToImage(synthesis,0.95);
}

PImage showLocalDisarrayPartiallyCoherent(float theReach, float theCoherence) {
  ArrayList<float[][]> tmpNX = new ArrayList<float[][]>();
  ArrayList<float[][]> tmpNY = new ArrayList<float[][]>();
  ArrayList<float[][]> tmp   = new ArrayList<float[][]>();
  for (int k=0; k<NUM_SCALES; k++) {
    float[][] layerX = new float[SZ][SZ];
    float[][] layerY = new float[SZ][SZ];
    for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) { 
      layerX[i][j] = theCoherence*coherentDataStackX.get(k)[i][j] + (1.0 - theCoherence)*randomGaussianScaledDataStackX.get(k)[i][j];
      layerY[i][j] = theCoherence*coherentDataStackY.get(k)[i][j] + (1.0 - theCoherence)*randomGaussianScaledDataStackY.get(k)[i][j];
    }
    tmpNX.add(layerX);
    tmpNY.add(layerY);
  }
  for (int k=0; k<NUM_SCALES; k++) tmp.add(disarray(tmpNX,tmpNY,k,theReach));
  float[][] synthesis = new float[SZ][SZ];
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) synthesis[i][j] = fiducialScaleLayers.get(NUM_SCALES-1)[i][j];
  for (int k=0; k<NUM_SCALES-1; k++) for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) synthesis[i][j] += tmp.get(k)[i][j];
  return dataToImage(synthesis,0.95);
}