import processing.core.*; 
import processing.data.*; 
import processing.event.*; 
import processing.opengl.*; 

import edu.emory.mathcs.jtransforms.fft.*; 
import edu.emory.mathcs.utils.*; 
import edu.emory.mathcs.jtransforms.dst.*; 
import edu.emory.mathcs.jtransforms.dht.*; 
import edu.emory.mathcs.jtransforms.dct.*; 
import java.util.Arrays; 
import java.util.concurrent.atomic.AtomicLong; 

import edu.emory.mathcs.jtransforms.fft.*; 
import edu.emory.mathcs.jtransforms.dht.*; 
import edu.emory.mathcs.jtransforms.dst.*; 
import edu.emory.mathcs.jtransforms.dct.*; 
import edu.emory.mathcs.utils.*; 

import java.util.HashMap; 
import java.util.ArrayList; 
import java.io.File; 
import java.io.BufferedReader; 
import java.io.PrintWriter; 
import java.io.InputStream; 
import java.io.OutputStream; 
import java.io.IOException; 

public class EidolonFactory extends PApplet {

String PATH = "Face.jpg"; // Default image
boolean imageHasBeenDefined = false;
boolean userStartedInterface = false;
boolean firsttime = false;

static final int SZ         = 512;
static final int NUM_SCALES = 14;

FloatFFT_2D fft;
Pcg32       rnd;
int         initialEpoch;

public void setup() {
   
  selectInput("Select a file to process:", "fileSelected");
  initialEpoch = millis();
  fft = new FloatFFT_2D(SZ,SZ);
  rnd = new Pcg32(); 
  font = loadFont("ComicSansMS-32.vlw");
  altFont = loadFont("Stencil-48.vlw");
  loadManualPages();
  // computeCorticalRepresentation(PATH);
  textFont(font, FONT_SIZE); 
  fill(FONT_COLOR); 
  noStroke(); 
  noLoop();
  surface.setTitle("Eidolon Factory");
}

public void fileSelected(File selection) {
  if (selection == null) {
    initializationReport = initializationReport+"\n"+"Computations for default image";
    redraw();
    computeCorticalRepresentation(PATH); // Use default image
    redraw();
  } 
  else {
    initializationReport = initializationReport+"\n"+"Computations for user selected image";
    computeCorticalRepresentation(selection.getAbsolutePath()); // User user selected image
    redraw();
  }
}

public void draw() {
  background(0);
  if (imageHasBeenDefined && userStartedInterface) {
    if (helpMode) {manual(-1,-1); helpMode=false;}
    else
      switch(mode) {
        case DISARRAY_RESULT: 
          if (displayMode==DEMO_MODE) {
            displayResults(resultDisplayMode); 
            resultsMenu(resultDisplayMode); 
          } 
          else manual(DISARRAY_RESULT,resultDisplayMode);
          break;
        case NOISE_FIELDS:
          if (displayMode==DEMO_MODE) {
            displayNoiseFields(theNoiseField);
            noiseFieldsMenu(theNoiseField); 
          } 
          else manual(NOISE_FIELDS,theNoiseField);
          break;
        case DATA_STACKS:
          if (displayMode==DEMO_MODE) {
            displayDataStack(theStack);
            dataStacksMenu(theStack); 
          } 
          else manual(DATA_STACKS,theStack);
          break;
      }
  }
  else {
    image(loadImage("Frontispiece.png"),0,0,2*SZ,SZ);
    fill(0,255,0);
    textFont(font,SMALL_FONT_SIZE);
    text(initializationReport,650,50);
    if (imageHasBeenDefined) {
      fill(255,255,0);
      textFont(font,SMALL_FONT_SIZE);
      text("ready when you are!",650,400);
      textFont(altFont,48);
      text("CLICKME!",650,450);
    }
    textFont(font,SMALL_FONT_SIZE);
  }
}

public void displayResults(int rdm) {
  int epoch = millis();
  fill(HEADING_COLOR);
  switch (rdm) {
    case FIDUCIAL_IMAGE:
      int lev = round(coherence*NUM_SCALES);
      if (lev<0) lev=0; else if (lev>=NUM_SCALES) lev = NUM_SCALES-1;
      image(captcha(fiducialImage, lev, reach),0,0,SZ,SZ);
      if (firsttime) {  
        fill(255,0,0);
        textFont(altFont,48);
        text("CLICKME!",100,100);
        firsttime = false;
      }
      fill(255,255,0);
      textFont(font,SMALL_FONT_SIZE);
      text("A.0. CAPTCHA of fiducial image",                                SZ+MARGIN, LINEHEIGHT);
      // text("Eidolon factory set up time "+(millis()-initialEpoch)+" ms", SZ+MARGIN, 2*LINEHEIGHT);
      break;
    case LOTZE_DISARRAY:
      lev = round(coherence*NUM_SCALES);
      if (lev<0) lev=0; else if (lev>=NUM_SCALES) lev = NUM_SCALES-1;
      computeRandomDisarrayStacks();
      theCurrentlyDisplayedImage = showLocalDisarrayLotze(lev,reach);
      image(theCurrentlyDisplayedImage,0,0,SZ,SZ);
      text("A.1. Lotze-type disarray ("+(millis()-epoch)+" ms)",              SZ+MARGIN, LINEHEIGHT);
      break;
   case HELMHOLTZ_DISARRAY:
      computeRandomScaledDisarrayStacks();
      theCurrentlyDisplayedImage = showLocalDisarrayHelmholtz(reach);
      image(theCurrentlyDisplayedImage,0,0,SZ,SZ);
      text("A.2. Helmholtz-type disarray ("+(millis()-epoch)+" ms)",           SZ+MARGIN, LINEHEIGHT);
      break;
    case COHERENT_DISARRAY:
      computeCoherentDisarrayStacks();
      computeCoherentDisarrayStacks();
      theCurrentlyDisplayedImage = showLocalDisarrayCoherent(reach);
      image(theCurrentlyDisplayedImage,0,0,SZ,SZ);
      text("A.3. Coherent disarray ("+(millis()-epoch)+" ms)",                 SZ+MARGIN, LINEHEIGHT);
      break;
    case PARTIALLY_COHERENT_DISARRAY:
      computeRandomScaledDisarrayStacks();
      computeCoherentDisarrayStacks();
      computeRandomScaledDisarrayStacks();
      theCurrentlyDisplayedImage = showLocalDisarrayPartiallyCoherent(reach,coherence);
      image(theCurrentlyDisplayedImage,0,0,SZ,SZ);
      text("A.4. Partially coherent disarray ("+(millis()-epoch)+" ms)",       SZ+MARGIN, LINEHEIGHT);
      break;
  }
  fill(255,0,0);
  text(fileSaveReport,50,50);
  fileSaveReport = "";
  fill(FONT_COLOR);
}

public void displayNoiseFields(int nf) {
  switch(nf){
    case LOTZE_STACK:
      theCurrentlyDisplayedImage = vectorImage(randomGaussianDataStackX.get(stackLevel),randomGaussianDataStackY.get(stackLevel));
      image(theCurrentlyDisplayedImage,0,0,SZ,SZ);
      image(noiseBinCounts(randomGaussianDataStackX.get(stackLevel),
                           randomGaussianDataStackY.get(stackLevel)),1.4f*SZ,0.5f*SZ,0.5f*SZ,0.3f*SZ);
      break;
    case HELMHOLTZ_STACK:
      theCurrentlyDisplayedImage = vectorImage(randomGaussianScaledDataStackX.get(stackLevel),randomGaussianScaledDataStackY.get(stackLevel));
      image(theCurrentlyDisplayedImage,0,0,SZ,SZ);
      image(noiseBinCounts(randomGaussianScaledDataStackX.get(stackLevel), 
                           randomGaussianScaledDataStackY.get(stackLevel)),1.4f*SZ,0.5f*SZ,0.5f*SZ,0.3f*SZ);
      break;
    case COHERENT_STACK:
      theCurrentlyDisplayedImage = vectorImage(coherentDataStackX.get(stackLevel),coherentDataStackY.get(stackLevel));
      image(theCurrentlyDisplayedImage,0,0,SZ,SZ);
      image(noiseBinCounts(coherentDataStackX.get(stackLevel),             
                           coherentDataStackY.get(stackLevel)),1.4f*SZ,0.5f*SZ,0.5f*SZ,0.3f*SZ);
      break;
  }
  textFont(font,SMALL_FONT_SIZE);
  text("Median "+nf(noiseQuantiles[0],3,3),SZ+MARGIN,0.5f*SZ-LINEHEIGHT);
  text("75% quantile "+nf(noiseQuantiles[1],3,3),SZ+MARGIN,0.5f*SZ-2*LINEHEIGHT);
  text("95% quantile "+nf(noiseQuantiles[2],3,3),SZ+MARGIN,0.5f*SZ-3*LINEHEIGHT);
  fill(255);
  text("reach histogram",1.1f*SZ,0.6f*SZ);
  textFont(font,FONT_SIZE);
}

public void displayDataStack(int ds) {
  switch(ds){
    case IMAGE_STACK:
      theCurrentlyDisplayedImage = dataToImage(fiducialScaleSpace.get(stackLevel),0.95f);
      image(theCurrentlyDisplayedImage,0,0,SZ,SZ);
      break;
    case MACCHIE:
      theCurrentlyDisplayedImage = macchie(stackLevel);
      image(theCurrentlyDisplayedImage,0,0,SZ,SZ);
      break;
    case LAPLACEAN_STACK:
      theCurrentlyDisplayedImage = dataToImage(fiducialScaleLayers.get(stackLevel),0.95f);
      image(theCurrentlyDisplayedImage,0,0,SZ,SZ);
      break;
    case EDGE_FINDER_STACK:
      theCurrentlyDisplayedImage = vectorImage(fiducialFirstOrderStackX.get(stackLevel),fiducialFirstOrderStackY.get(stackLevel));
      image(theCurrentlyDisplayedImage,0,0,SZ,SZ);
      break;
    case EDGELETS:
      theCurrentlyDisplayedImage = edgelets(stackLevel);
      image(theCurrentlyDisplayedImage,0,0,SZ,SZ);
      break;
    case LINE_FINDER_STACK:
      theCurrentlyDisplayedImage = lineFinderBasisImage(fiducialSecondOrderStackP.get(stackLevel),fiducialSecondOrderStackQ.get(stackLevel),fiducialSecondOrderStackR.get(stackLevel));
      image(theCurrentlyDisplayedImage,0,0,SZ,SZ);
      break;
  }
}
static final int PERIODIC_BOUNDARY_CONDITIONS = 0;
static final int CLAMPED_BOUNDARY_CONDITIONS  = 1;
static final int CONSTANT_BOUNDARY_CONDITIONS = 2;
static final int BOUNDARY_CONDITIONS = PERIODIC_BOUNDARY_CONDITIONS;

public PImage sanityCheck() {
  float[][] synthesis = new float[SZ][SZ];
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) synthesis[i][j] = fiducialScaleLayers.get(NUM_SCALES-1)[i][j];
  for (int k=0; k<NUM_SCALES-1; k++) for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) synthesis[i][j] += fiducialScaleLayers.get(k)[i][j];
  return dataToImage(synthesis);
}

public int periodic(int i, int n) { return ((i % n) + n) % n; }

public boolean inRange(int i) {
  return ((i>=0) && (i<SZ));
}

public boolean inRange2D( int i, int j) {
  return (inRange(i) && inRange(j));
}

// Disarrays an image
public PImage captcha(PImage anImage, int theLevel, float theReach) {
  float[][] data = imageToData(anImage);
  float[][] disarrayed = new float[SZ][SZ];
  ArrayList<float[][]> perturbation = twoRandomFields(theLevel);
  float[][] xComponent = perturbation.get(0);
  float[][] yComponent = perturbation.get(1);
  switch (BOUNDARY_CONDITIONS) {
    case PERIODIC_BOUNDARY_CONDITIONS:
      for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) 
          disarrayed[i][j] = data[periodic((int) (i + theReach*xComponent[i][j]),SZ)][periodic((int) (j + theReach*yComponent[i][j]),SZ)];
      break;
    case CLAMPED_BOUNDARY_CONDITIONS:
      for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) 
          disarrayed[i][j] = data[(int) constrain(i + theReach*xComponent[i][j],0,SZ-1)][(int) constrain(j + theReach*yComponent[i][j],0,SZ-1)];
      break;
    case CONSTANT_BOUNDARY_CONDITIONS:
      for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) {
        int ii = (int) (i + theReach*xComponent[i][j]);
        int jj = (int) (j + theReach*yComponent[i][j]);
        if (inRange2D(i,j)) disarrayed[i][j] = fiducialScaleLayers.get(theLevel)[ii][jj];
        else disarrayed[i][j] = 127.0f;
      }
      break;
  }
  return dataToImage(disarrayed);
}

// Disarrays a single data plane
public float[][] disarray(ArrayList<float[][]> theStackX, ArrayList<float[][]> theStackY, int theLevel, float theReach) {
  
  float[][] disarrayed = new float[SZ][SZ];
  
  switch (BOUNDARY_CONDITIONS) {
    
    case PERIODIC_BOUNDARY_CONDITIONS:
      for (int i=0; i<SZ; i++) 
        for (int j=0; j<SZ; j++) 
          disarrayed[i][j] = 
            fiducialScaleLayers.get(theLevel)[
                periodic((int) (i + theReach*theStackX.get(theLevel)[i][j]),SZ)]
              [
                periodic((int) (j + theReach*theStackY.get(theLevel)[i][j]),SZ)
              ];
      break;
      
    case CLAMPED_BOUNDARY_CONDITIONS:
      for (int i=0; i<SZ; i++) 
        for (int j=0; j<SZ; j++) 
          disarrayed[i][j] = 
            fiducialScaleLayers.get(theLevel)[
              (int) constrain(i + theReach*theStackX.get(theLevel)[i][j],0,SZ-1)
            ][
              (int) constrain(j + theReach*theStackY.get(theLevel)[i][j],0,SZ-1)
            ];
      break;
      
    case CONSTANT_BOUNDARY_CONDITIONS:
      for (int i=0; i<SZ; i++) 
        for (int j=0; j<SZ; j++) {
        int ii = (int) (i + theReach*theStackX.get(theLevel)[i][j]);
        int jj = (int) (j + theReach*theStackY.get(theLevel)[i][j]);
        if (inRange2D(i,j))
          disarrayed[i][j] = fiducialScaleLayers.get(theLevel)[ii][jj];
        else
          disarrayed[i][j] = 127.0f;
      }
      break;
  }
  return disarrayed;
}

// Shows the effect of local disarray for a Lotze noise stack
public PImage showLocalDisarrayLotze(int theLevel, float theReach) {
  
  ArrayList<float[][]> noiseStackX = new ArrayList<float[][]>();
  ArrayList<float[][]> noiseStackY = new ArrayList<float[][]>();
  for (int k=0; k<NUM_SCALES; k++) {
    ArrayList<float[][]> tmp = twoRandomFields(theLevel);
    noiseStackX.add(tmp.get(0));
    noiseStackY.add(tmp.get(1));
  }
  
  ArrayList<float[][]> tmp = new ArrayList<float[][]>();
  for (int k=0; k<NUM_SCALES; k++) tmp.add(disarray(noiseStackX,noiseStackY,k,theReach));
  float[][] synthesis = new float[SZ][SZ];
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) synthesis[i][j] = fiducialScaleLayers.get(NUM_SCALES-1)[i][j];
  for (int k=0; k<NUM_SCALES-1; k++) for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) synthesis[i][j] += tmp.get(k)[i][j];
  return dataToImage(synthesis);
}

public PImage showLocalDisarrayHelmholtz(float theReach) {
  ArrayList<float[][]> tmp = new ArrayList<float[][]>();
  for (int k=0; k<NUM_SCALES; k++) tmp.add(disarray(randomGaussianScaledDataStackX,randomGaussianScaledDataStackY,k,theReach));
  float[][] synthesis = new float[SZ][SZ];
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) synthesis[i][j] = fiducialScaleLayers.get(NUM_SCALES-1)[i][j];
  for (int k=0; k<NUM_SCALES-1; k++) for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) synthesis[i][j] += tmp.get(k)[i][j];
  return dataToImage(synthesis);
}

public PImage showLocalDisarrayCoherent(float theReach) {
  ArrayList<float[][]> tmp = new ArrayList<float[][]>();
  for (int k=0; k<NUM_SCALES; k++) tmp.add(disarray(coherentDataStackX,coherentDataStackY,k,theReach));
  float[][] synthesis = new float[SZ][SZ];
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) synthesis[i][j] = fiducialScaleLayers.get(NUM_SCALES-1)[i][j];
  for (int k=0; k<NUM_SCALES-1; k++) for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) synthesis[i][j] += tmp.get(k)[i][j];
  return dataToImage(synthesis,0.95f);
}

public PImage showLocalDisarrayPartiallyCoherent(float theReach, float theCoherence) {
  ArrayList<float[][]> tmpNX = new ArrayList<float[][]>();
  ArrayList<float[][]> tmpNY = new ArrayList<float[][]>();
  ArrayList<float[][]> tmp   = new ArrayList<float[][]>();
  for (int k=0; k<NUM_SCALES; k++) {
    float[][] layerX = new float[SZ][SZ];
    float[][] layerY = new float[SZ][SZ];
    for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) { 
      layerX[i][j] = theCoherence*coherentDataStackX.get(k)[i][j] + (1.0f - theCoherence)*randomGaussianScaledDataStackX.get(k)[i][j];
      layerY[i][j] = theCoherence*coherentDataStackY.get(k)[i][j] + (1.0f - theCoherence)*randomGaussianScaledDataStackY.get(k)[i][j];
    }
    tmpNX.add(layerX);
    tmpNY.add(layerY);
  }
  for (int k=0; k<NUM_SCALES; k++) tmp.add(disarray(tmpNX,tmpNY,k,theReach));
  float[][] synthesis = new float[SZ][SZ];
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) synthesis[i][j] = fiducialScaleLayers.get(NUM_SCALES-1)[i][j];
  for (int k=0; k<NUM_SCALES-1; k++) for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) synthesis[i][j] += tmp.get(k)[i][j];
  return dataToImage(synthesis,0.95f);
}
/* ************************************************************************ */
/*                                                                          */
/*                UTILITIES FOR FAST FOURIER TRANSFORM METHODS              */   
/*                                                                          */
/* ************************************************************************ */

// IMPORT JTRANSFORMS (IN code DIRECTORY OF SKETCH






// Convolution
public float[][] runConvolution(float[][] imageSpectrum, float[][] kernelSpectrum){
  // Find the complexproduct of the spectra
  float[][] tmp = complexMultiply(imageSpectrum,kernelSpectrum);
  // The inverse Fourier transform yields the filtered image
  fft.complexInverse(tmp,true);
  return getRealPart(tmp);
}

// Given a float data buffer create a spectrum
public float[][] imageSpectrum(float[][] data){
  // Compute the Fourier spectrum of the image. This yields a complex spectrum. Re and Im parts are interlaced!
  float[][] spectrum = new float[SZ][2*SZ];
  for (int i=0; i<SZ; i++) for(int j=0; j<SZ; j++) {
    spectrum[i][j]    = data[i][j]; // Re
    spectrum[i][j+SZ] =        0.0f; // Im
  }
  fft.realForwardFull(spectrum); // spectrum contains the image spectrum
  return spectrum;
}

public float[][] kernelSpectrum(float sigma){
 float[][] kernel = new float[SZ][SZ];
 for (int i = 0; i < SZ; i++) for (int j = 0; j < SZ; j++) {
  float u = (float) i;
  float v = (float) j;
  if (i > SZ/2) u = u - SZ;
  if (j > SZ/2) v = v - SZ;
  kernel[i][j] = gaussian2D(u,v,sigma);
 }
 float[][] spectrum = new float[SZ][2*SZ];
 for (int i=0; i<SZ; i++) for(int j=0; j<SZ; j++) { 
   spectrum[i][j]     = kernel[i][j]; // Re
   spectrum[i][j+SZ] = 0.0f;           // Im
 }
 // The transform is complex, Re and Im interlaced
 fft.realForwardFull(spectrum);
 return spectrum;
}

public float[][] kernelSpectrum(int xOrder, int yOrder, float sigma){
  // Construct the kernel
  // This is a real function
  float[][] kernel = new float[SZ][SZ];
  float[]   column = new float[SZ];
  float[]      row = new float[SZ];
  for (int i=0; i<SZ; i++) {
    float u = (float) i;
    if (i > SZ/2) u = u - SZ;
    column[i] = pow( -1.0f/(sigma*sqrt(2.0f)),xOrder) * hermitePolynomial(xOrder,u/(sigma*sqrt(2))) * gaussian(u,sigma);
  }
  for(int j=0; j<SZ; j++) {
    float v = (float) j;
    if (j > SZ/2) v = v - SZ;
    row[j] = pow( -1.0f/(sigma*sqrt(2.0f)),yOrder) * hermitePolynomial(yOrder,v/(sigma*sqrt(2))) * gaussian(v,sigma);
  }
  for (int i = 0; i < SZ; i++) for (int j = 0; j < SZ; j++) kernel[i][j] = column[i] * row[j];
  // Fourier transform it, creating a new array if none exists, otherwise overwrite
  // The transform is complex, Re and Im interlaced
  // The kernel itself is simply filled up with zeroes
  float[][] spectrum = new float[SZ][2*SZ];
  for (int i=0; i<SZ; i++) for(int j=0; j<SZ; j++) { 
      spectrum[i][j]    = kernel[i][j]; // Re
      spectrum[i][j+SZ] = 0.0f;          // Im
    }
  fft.realForwardFull(spectrum);
  return spectrum;
}

// Returns the product of two complex spectra a & b
// Re and Im are interlaced!
public float[][] complexMultiply(float[][] a, float[][] b) {
  float[][] result = new float[SZ][2*SZ];
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) {
    result[i][2*j]   = a[i][2*j] * b[i][  2*j] - a[i][2*j+1] * b[i][2*j+1];
    result[i][2*j+1] = a[i][2*j] * b[i][2*j+1] + a[i][2*j+1] * b[i][  2*j];
  }
  return result;
}

// This picks out the real part and ignores the imaginary part
// Re and Im are interlaced!
public float[][] getRealPart(float[][] aComplexSpectrum) {
  float[][] result = new float[SZ][SZ];
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) result[i][j] = aComplexSpectrum[i][2*j];
  return result;
}

// 1D Gaussian
public float gaussian(float x, float sigma) {return exp(-sq(x)/(2f*sq(sigma)))/(sqrt(2f*PI)*sigma);}

// 2D Gaussian
public float gaussian2D(float x, float y, float sigma) {return exp(-(sq(x)+sq(y))/(2f*sq(sigma)))/(2f*PI*sq(sigma));}

// Hermite polynomials, via the recursion relation
public float hermitePolynomial(int n, float x) {
  float result;
  switch (n) {
    case(0):  result = 1.0f;
              break;
    case(1):  result = 2.0f * x;
              break;
    case(2):  result = 4.0f * sq(x) - 2.0f;
              break;
    case(3):  result = 8.0f * pow(x,3) - 12.0f * x;
              break;
    case(4):  result = 16.0f * pow(x,4) - 48.0f * sq(x) + 12.0f;
              break;
    default:  result = 2.0f*(x*hermitePolynomial(n-1,x)-(n-1)*hermitePolynomial(n-2,x));
  }
  return result;
}


static final int NOISE_NUM_BINS = 100;
static final float NOISE_CUTOFF = 0.99f;
float[] noiseQuantiles = new float[3];

// Extract the green channel of an RGB image as a single float data plane
public float[][] imageToData(PImage im) {
 float[][] data = new float[SZ][SZ];
 im.loadPixels();
 for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) data[i][j] = green(im.pixels[j*SZ+i]);
 return data;
}

// Put the data in all color channels of an RGB image
public PImage dataToImage(float[][] data) {
 PImage im = new PImage(SZ,SZ);
 im.loadPixels();
 for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) im.pixels[j*SZ+i] = color(data[i][j],data[i][j],data[i][j]);
 im.updatePixels();
 return im;
}

public float[][] invertData(float[][] data) {
  float[][] invData = new float[SZ][SZ];
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) invData[i][j] = -data[i][j];
  return invData;
}

// Given float data create a monochrome image
// level is a quantile limit, useful to set the scale
// For level try 0.99, 0.95, or 0.9 say
public PImage dataToImage(float[][] buffer, float level){
  float p = quantile(buffer,level);
  float q = quantile(buffer,1-level);
  // Create an image
  PImage im = new PImage(SZ,SZ);
  // Fill a monochrome image using levels between 1-level and level:
  im.loadPixels();
  float tmp = 255.0f/(q-p);
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) {
      float val = constrain(tmp*(buffer[i][j]-p),0.0f,255.0f);
      im.pixels[j*SZ+i] = color(255-val);
    }
  im.updatePixels();
  return im;
}

// Given float data create a monochrome image
// level are the quartiles and 5% limits, useful to set the scale
public PImage dataToPoster(float[][] data){
  float[] p = quartilesAnd5PercentLevels(data);
  // Create an image
  PImage im = new PImage(SZ,SZ);
  // Fill a monochrome image using all available levels
  im.loadPixels();
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) 
    if      (data[i][j]<p[0]) im.pixels[j*SZ+i] = color(  0);
    else if (data[i][j]<p[1]) im.pixels[j*SZ+i] = color( 63);
    else if (data[i][j]<p[2]) im.pixels[j*SZ+i] = color(127); 
    else if (data[i][j]<p[3]) im.pixels[j*SZ+i] = color(191);
    else                      im.pixels[j*SZ+i] = color(255);
  im.updatePixels();
  return im;
}

public PImage macchie(int theLevel) {
  float[][] data = fiducialScaleSpace.get(theLevel);
  float[] tmp = flattenData(data);
  Arrays.sort(tmp);
  float darkToShade      = tmp[round(    tmp.length/6.0f)];
  float shadeToNeutral   = tmp[round(2.0f*tmp.length/6.0f)];
  float neutralToLight   = tmp[round(4.0f*tmp.length/6.0f)];
  float lightToHighlight = tmp[round(5.0f*tmp.length/6.0f)];
  PImage posterized = new PImage(SZ,SZ);
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) {
    float val = data[i][j];
    int c = color(127);
    if      (val<darkToShade)                              c = color(0);
    else if ((val<shadeToNeutral) && (val>darkToShade))    c = color(63);
    else if ((val<neutralToLight) && (val>shadeToNeutral)) c = color(196);
    else if (val>lightToHighlight)                         c = color(255);
    posterized.set(i,j,c);
  }
  return posterized;
}

public PImage edgelets(int theLevel) {
  float[][] data = fiducialScaleLayers.get(theLevel);
  float[] tmp = flattenData(data);
  Arrays.sort(tmp);
  float darkToShade      = tmp[round(0.1f*tmp.length)];
  float shadeToNeutral   = tmp[round(0.2f*tmp.length)];
  float neutralToLight   = tmp[round(0.8f*tmp.length)];
  float lightToHighlight = tmp[round(0.9f*tmp.length)];
  PImage posterized = new PImage(SZ,SZ);
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) {
    float val = data[i][j];
    int c = color(127);
    if      (val<darkToShade)                              c = color(0);
    else if ((val<shadeToNeutral) && (val>darkToShade))    c = color(63);
    else if ((val<neutralToLight) && (val>shadeToNeutral)) c = color(196);
    else if (val>lightToHighlight)                         c = color(255);
    posterized.set(i,j,c);
  }
  return posterized;
}

public float[] flattenData(float[][] data) {
  float[] flat = new float[SZ*SZ];
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) flat[j*SZ+i] = data[i][j];
  return flat;
}

public float[] meanAndStandarddeviation(float[][] data) {
  float[] tmp = flattenData(data);
  float s  = 0.0f;
  float ss = 0.0f;
  for (int i=0; i<SZ*SZ; i++) {
    s  +=    tmp[i];
    ss += sq(tmp[i]);
  }
  // Find mean and standard deviations
  s  /= (SZ*SZ);
  ss /= (SZ*SZ);
  ss = sqrt(abs(sq(s)-ss));
  float[] result = {s,ss};
  return result;
}

public int[] binCounts(float[][] data, float mn, float mx, int numBins) {
  float[] tmp = flattenData(data);
  float rn = mx- mn;
  int[] counts = new int[numBins];
  int j;
  for (int i=0;i<SZ*SZ; i++) {
    j = floor(numBins*(tmp[i]-mn)/rn);
    if ((j>=0)&&(j<numBins)) counts[j]++;
  }
  return counts;
}

public PImage imageHistogram(float[][] data, int numBins) {
  int[] counts = binCounts(data, 0.0f, 255.0f, numBins);
  float[] fc = new float[numBins];
  float tmp = 0.0f;
  for (int i=0; i<numBins; i++) if (counts[i]>tmp) tmp=counts[i];
  for (int i=0; i<numBins; i++) fc[i] = counts[i]/tmp;
  PGraphics g = createGraphics(SZ,SZ,JAVA2D);
  g.beginDraw();
    g.background(127);
    g.stroke(0);
    g.strokeWeight(0.5f);
    g.fill(127,127,255);
    g.rect(50,50,SZ-100,SZ-100);
    g.noStroke();
    float  w = (SZ - 100.0f)/numBins;
    for (int i=0; i<numBins; i++) {
      g.fill(255.0f*(i+0.5f)/numBins);
      g.rect(50+i*w,SZ-50-fc[i]*(SZ-150),w,fc[i]*(SZ-150));
    }
  g.endDraw();
  return g;
}

public PImage noiseBinCounts(float[][] layerX, float[][] layerY){
  FloatList d = new FloatList();
  for (int k=0; k<NUM_SCALES; k++) 
    for (int i=0; i<SZ; i++) 
      for (int j=0; j<SZ; j++) 
        d.append(sqrt(sq(layerX[i][j]) + sq(layerY[i][j])));
  d.sort();
  float cutoff = d.get(round(NOISE_CUTOFF*d.size()));
  noiseQuantiles[0] = d.get(round(0.50f*d.size()));
  noiseQuantiles[1] = d.get(round(0.75f*d.size()));
  noiseQuantiles[2] = d.get(round(0.95f*d.size()));
  int numBins = 100;
  int[] counts = new int[numBins];
  int j;
  for (int i=0;i<d.size(); i++) {
    j = floor(NOISE_NUM_BINS*(d.get(i)/cutoff));
    if ((j>=0)&&(j<NOISE_NUM_BINS)) counts[j]++;
  }
  int top = 0;
  for (int i=0;i<NOISE_NUM_BINS; i++) if (counts[i]>top) top=counts[i];
  float[] tmp = new float[NOISE_NUM_BINS];
  for (int i=0;i<NOISE_NUM_BINS; i++) tmp[i] = PApplet.parseFloat(counts[i])/top;
  PGraphics g = createGraphics(SZ,SZ,JAVA2D);
  g.beginDraw();
    g.background(255);
    g.noStroke();
    g.fill(0);
    for (int i=0;i<NOISE_NUM_BINS; i++) g.rect((i*SZ/NOISE_NUM_BINS),(1-tmp[i])*SZ,SZ/NOISE_NUM_BINS,tmp[i]*SZ);
    g.noFill();
    g.stroke(255,255,0);
    g.strokeWeight(8.0f);
    g.rect(0,0,SZ,SZ);
  g.endDraw();
  return g;
}

// Return the quartiles of data
public float[] quartilesAnd5PercentLevels(float[][] data) {
  float[] tmp = flattenData(data);
  Arrays.sort(tmp);
  float[] q = {tmp[round(0.05f*tmp.length)],tmp[round(0.25f*tmp.length)],tmp[round(0.5f*tmp.length)],tmp[round(0.75f*tmp.length)],tmp[round(0.95f*tmp.length)],};
  return q;
}

// Return a quantile of data
public float quantile(float[][] data, float theLevel) {
  float[] tmp = flattenData(data);
  Arrays.sort(tmp);
  return(tmp[round(theLevel*tmp.length)]);
}

// Planar vector image - hue drom direction, saturation signifies magnitude
public PImage vectorImage(float[][] x, float[][] y) {
  float[][] rho = new float[SZ][SZ];
  float[][] phi = new float[SZ][SZ];
  float maxRho = 0.0f;
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++)  {
    rho[i][j] = sq(x[i][j])+sq(y[i][j]);
    if (rho[i][j]>maxRho) maxRho = rho[i][j];
    phi[i][j] = atan2(y[i][j],x[i][j]);
  }
  float limit = 0.9f; // the maximum saturation
  float cnst = maxRho*(1.0f-limit)/limit;
  PImage im = new PImage(SZ,SZ);
  im.loadPixels();
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) 
    im.pixels[j*SZ+i] = lerpColor(color(255,255,255),directionColor(phi[i][j]),rho[i][j]/(cnst+rho[i][j]));
  im.updatePixels();
  return im;
}

// Shows the 3D line finder vector basis in RGB
public PImage lineFinderBasisImage(float[][] secondOrderP,float[][] secondOrderQ, float[][] secondOrderR) {
  PImage im = new PImage(SZ,SZ);
  float factor = 127.0f/max(quantile(secondOrderP,0.75f), quantile(secondOrderQ,0.75f), quantile(secondOrderR,0.75f));
  im.loadPixels();
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) 
    im.pixels[j*SZ+i] = color(127+factor*secondOrderP[i][j],127+factor*secondOrderQ[i][j],127+factor*secondOrderR[i][j]);
  im.updatePixels();
  return im;
}

public PImage lineFinderBasisColorScale() {
  PGraphics g = createGraphics(SZ,SZ,JAVA2D);
  g.beginDraw();
    g.background(0);
    g.ellipseMode(RADIUS);
    g.noStroke();
    float semiSector = PI/6.0f;
    float phi0 = 0.0f;
    float phi1 = PI/3.0f;
    float phi2 = 2.0f*PI/3.0f;
    g.fill(255,0,0);
    g.arc(0.5f*SZ,0.5f*SZ,0.5f*SZ,0.5f*SZ,phi0-semiSector,phi0+semiSector,PIE);
    g.arc(0.5f*SZ,0.5f*SZ,0.5f*SZ,0.5f*SZ,phi0+PI-semiSector,phi0+PI+semiSector,PIE);
    g.fill(0,255,0);
    g.arc(0.5f*SZ,0.5f*SZ,0.5f*SZ,0.5f*SZ,phi1-semiSector,phi1+semiSector,PIE);
    g.arc(0.5f*SZ,0.5f*SZ,0.5f*SZ,0.5f*SZ,phi1+PI-semiSector,phi1+PI+semiSector,PIE);
    g.fill(0,0,255);
    g.arc(0.5f*SZ,0.5f*SZ,0.5f*SZ,0.5f*SZ,phi2-semiSector,phi2+semiSector,PIE);
    g.arc(0.5f*SZ,0.5f*SZ,0.5f*SZ,0.5f*SZ,phi2+PI-semiSector,phi2+PI+semiSector,PIE);
    g.fill(255,127);
    g.ellipse(0.5f*SZ,0.5f*SZ,SZ/3.0f,SZ/3.0f);
    g.fill(255,196);
    g.ellipse(0.5f*SZ,0.5f*SZ,SZ/6.0f,SZ/6.0f);
  g.endDraw();
  return g;
}

// Shows a direction image
public PImage directionImage(float[][] direction) {
  PImage im = new PImage(SZ,SZ);
  im.loadPixels();
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) 
    im.pixels[j*SZ+i] = directionColor(direction[i][j]);
  im.updatePixels();
  return im;
}

public int directionColor(float direction) {
  direction = (1+direction/PI)*3.0f; // From (-PI,+PI) to (0,6)
  int   i = floor(direction);
  float f = direction - i;
  int result = color(127);
  switch (i) {
    case 0: result = lerpColor(color(255, 255,   0),color(  0, 255,   0),f); break; // YELLOW  to GREEN
    case 1: result = lerpColor(color(0,   255,   0),color(  0, 255, 255),f); break; // GREEN   to CYAN
    case 2: result = lerpColor(color(0,   255, 255),color(  0,   0, 255),f); break; // CYAN    to BLUE
    case 3: result = lerpColor(color(0,     0, 255),color(255,   0, 255),f); break; // BLUE    to MAGENTA
    case 4: result = lerpColor(color(255,   0, 255),color(255,   0,   0),f); break; // MAGENTA to RED
    case 5: result = lerpColor(color(255,   0,   0),color(255, 255,   0),f); break; // RED     to YELLOW
    default: break;
  }
  return(result);
}

public PImage directionColorScale() {
  PGraphics g = createGraphics(SZ,SZ,JAVA2D);
  g.beginDraw();
    g.background(0);
    g.ellipseMode(RADIUS);
    g.noStroke();
    for (int i=0; i<36; i++) {
      float phi1 = i*(TWO_PI/36.0f)-PI;
      float phi2 = (i+1.0f)*(TWO_PI/36.0f)-PI;
      g.fill(directionColor(0.5f*(phi1+phi2)));
      g.arc(0.5f*SZ,0.5f*SZ,0.5f*SZ,0.5f*SZ,phi1,phi2,PIE);
    }
    g.fill(255,127);
    g.ellipse(0.5f*SZ,0.5f*SZ,SZ/3.0f,SZ/3.0f);
    g.fill(255,196);
    g.ellipse(0.5f*SZ,0.5f*SZ,SZ/6.0f,SZ/6.0f);
  g.endDraw();
  return g;
}
static final int DOWN_KEY  = 0;
static final int UP_KEY    = 1;
static final int LEFT_KEY  = 2;
static final int RIGHT_KEY = 3;

static final float REACH_FACTOR = 1.05f;
static final float MIN_REACH    = 0.05f;
static final float MAX_REACH    = 10.0f;

static final float COHERENCE_STEP = 0.05f;
static final float MIN_COHERENCE  = 0.01f;
static final float MAX_COHERENCE  = 0.99f;

static final int LOTZE_STACK      = 0;
static final int HELMHOLTZ_STACK  = 1;
static final int COHERENT_STACK   = 2;
static final int NUM_NOISE_FIELDS = 3;
int theNoiseField = LOTZE_STACK;

static final int DISARRAY_RESULT = 0;
static final int NOISE_FIELDS    = 1;
static final int DATA_STACKS     = 2;
static final int NUM_MODES       = 3;
int mode = DISARRAY_RESULT;

static final int IMAGE_STACK       = 0;
static final int MACCHIE           = 1;
static final int LAPLACEAN_STACK   = 2;
static final int EDGE_FINDER_STACK = 3;
static final int EDGELETS          = 4;
static final int LINE_FINDER_STACK = 5;
static final int NUM_STACKS        = 6;
int theStack = IMAGE_STACK;

static final int FIDUCIAL_IMAGE              = 0;
static final int LOTZE_DISARRAY              = 1;
static final int HELMHOLTZ_DISARRAY          = 2;
static final int COHERENT_DISARRAY           = 3;
static final int PARTIALLY_COHERENT_DISARRAY = 4;
static final int NUM_RESULT_DISPLAY_MODES    = 5;
int resultDisplayMode = FIDUCIAL_IMAGE;

static final int DEMO_MODE = 0;
static final int TEXT_MODE = 1;
int displayMode = DEMO_MODE;

PFont font, altFont;
static final int MARGIN            = 10;
static final int LINEHEIGHT        = 30;
int              FONT_SIZE         = ceil(SZ/22);
int              SMALL_FONT_SIZE   = ceil(SZ/32);
int            FONT_COLOR        = color(0,255,0);
int            BOTTOM_LINE_COLOR = color(255,127,0);
int            HEADING_COLOR     = color(255,255,0);

float reach      = 1.0f;
float coherence  = 0.5f;
int   stackLevel = 0;

PImage theCurrentlyDisplayedImage; 

String fileSaveReport = "";

ArrayList<PImage> HELP = new ArrayList<PImage>();
ArrayList<PImage> A_0 = new ArrayList<PImage>();
ArrayList<PImage> A_1 = new ArrayList<PImage>();
ArrayList<PImage> A_2 = new ArrayList<PImage>();
ArrayList<PImage> A_3 = new ArrayList<PImage>();
ArrayList<PImage> A_4 = new ArrayList<PImage>();
ArrayList<PImage> B_0 = new ArrayList<PImage>();
ArrayList<PImage> B_1 = new ArrayList<PImage>();
ArrayList<PImage> B_2 = new ArrayList<PImage>();
ArrayList<PImage> C_0 = new ArrayList<PImage>();
ArrayList<PImage> C_1 = new ArrayList<PImage>();
ArrayList<PImage> C_2 = new ArrayList<PImage>();
ArrayList<PImage> C_3 = new ArrayList<PImage>();
ArrayList<PImage> C_4 = new ArrayList<PImage>();
ArrayList<PImage> C_5 = new ArrayList<PImage>();

boolean helpMode = false;

int manualPage = 0;

public void mouseClicked() {
  if (imageHasBeenDefined) {
    userStartedInterface = true;
    redraw();
  }
  if (userStartedInterface) {
    if      (displayMode == DEMO_MODE && mouseX<SZ) {
      displayMode = TEXT_MODE;
      manualPage = 0;
    }
    else if (displayMode == TEXT_MODE) displayMode = DEMO_MODE;
  }
}

public void keyPressed(){
  if (key==CODED) {
    if (displayMode == DEMO_MODE) {
      if      (keyCode== DOWN) handleArrowKeys(0);
      else if (keyCode==   UP) handleArrowKeys(1);
      else if (keyCode== LEFT) handleArrowKeys(2);
      else if (keyCode==RIGHT) handleArrowKeys(3);
    }
    else {
      if      (keyCode== LEFT) { manualPage--; redraw();}
      else if (keyCode==RIGHT) { manualPage++; redraw();}
    }
  }
  else if (key==' ') redraw();
  else if (key=='0') handleNumericKeys(0);
  else if (key=='1') handleNumericKeys(1);
  else if (key=='2') handleNumericKeys(2);
  else if (key=='3') handleNumericKeys(3);
  else if (key=='4') handleNumericKeys(4);
  else if (key=='5') handleNumericKeys(5);
  else if (key=='6') handleNumericKeys(6);
  else if ((key=='a')||(key=='A')) handleLetterKeys(0);
  else if ((key=='b')||(key=='B')) handleLetterKeys(1);
  else if ((key=='c')||(key=='C')) handleLetterKeys(2);
  else if ((key=='h')||(key=='H')) { helpMode = true; redraw();}
  else if ((key==RETURN)||(key==ENTER)) selectOutput("Select a file to write to:", "outputFileSelected");
}

public void outputFileSelected(File selection) {
  if (selection == null) {
    fileSaveReport = "File selection failed!";
  } 
  else {
    theCurrentlyDisplayedImage.save(selection.getAbsolutePath());
    fileSaveReport = "File saved!";
  }
  redraw();
}

public void handleArrowKeys(int theKey) {
  switch (mode) {
    case DISARRAY_RESULT:
      switch (theKey) {
        case DOWN_KEY:
          reach /= REACH_FACTOR;
          if (reach<0.001f) reach = MIN_REACH;
          redraw();
          break;
        case UP_KEY:
          reach *= REACH_FACTOR;
          if (reach>1000.0f) reach = MAX_REACH;
          redraw();
          break;
        case LEFT_KEY:
          coherence -= COHERENCE_STEP;
          if (coherence<0.01f) coherence = MIN_COHERENCE;
          redraw();
          break;
        case RIGHT_KEY:
          coherence += COHERENCE_STEP;
          if (coherence>0.99f) coherence = MAX_COHERENCE;
          redraw();
          break;
      }
      break;
    case NOISE_FIELDS:
    case DATA_STACKS:
      switch (theKey) {
        case DOWN_KEY:
          stackLevel--;
          if (stackLevel<0) stackLevel = 0;
          redraw();
          break;
        case UP_KEY:
          stackLevel++;
          if (stackLevel>=NUM_SCALES) stackLevel = NUM_SCALES-1;
          redraw();
          break;
      }
      break;
  }
}

public void handleLetterKeys(int i){
  mode = i;
  if      (mode<0)          mode = DISARRAY_RESULT;
  else if (mode>=NUM_MODES) mode = DATA_STACKS;
  switch (mode) {
    case DISARRAY_RESULT: { resultDisplayMode = FIDUCIAL_IMAGE; break;}
    case NOISE_FIELDS   : { theNoiseField     = LOTZE_STACK;    break;}
    case DATA_STACKS    : { theStack          = IMAGE_STACK;    break;}
  }
  redraw();
}

public void handleNumericKeys(int i) {
  switch (mode) {
    case DISARRAY_RESULT:
      resultDisplayMode = i;
      if (resultDisplayMode<0) resultDisplayMode=FIDUCIAL_IMAGE; 
      else if (resultDisplayMode>=NUM_RESULT_DISPLAY_MODES) resultDisplayMode = PARTIALLY_COHERENT_DISARRAY;
      redraw();
    case NOISE_FIELDS:
      theNoiseField = i;
      if (theNoiseField<0) theNoiseField=LOTZE_STACK; 
      else if (theNoiseField>=NUM_NOISE_FIELDS) theNoiseField = COHERENT_STACK;
      redraw();
      break;
    case DATA_STACKS:
      theStack = i;
      if (theStack<0) theStack=IMAGE_STACK; 
      else if (theStack>=NUM_STACKS) theStack = LINE_FINDER_STACK;
      redraw();
      break;
  }
}

public void resultsMenu(int rdm) {
  switch (rdm) {
    case FIDUCIAL_IMAGE:
      textFont(font,SMALL_FONT_SIZE);
      text("UP-DOWN arrow keys to adjust reach",         SZ+MARGIN, 7*LINEHEIGHT);
      text("Reach = "+nf(reach,3,3),                     SZ+MARGIN, 8*LINEHEIGHT);
      text("LEFT-RIGHT keys to adjust grain-size",  SZ+MARGIN, 9*LINEHEIGHT);
      int lev = round(coherence*NUM_SCALES);
      if (lev<0) lev=0; else if (lev>=NUM_SCALES) lev = NUM_SCALES-1;
      text("Grain-size = "+nf(scales[lev],3,3), SZ+MARGIN, 10*LINEHEIGHT);
      break;
    case LOTZE_DISARRAY:
    case HELMHOLTZ_DISARRAY:
    case COHERENT_DISARRAY:
      text("UP-DOWN arrow keys to adjust reach",         SZ+MARGIN, 7*LINEHEIGHT);
      text("Reach = "+nf(reach,3,3),                     SZ+MARGIN, 8*LINEHEIGHT);
      text("LEFT-RIGHT keys to adjust grain-size",       SZ+MARGIN, 9*LINEHEIGHT);
      lev = round(coherence*NUM_SCALES);
      if (lev<0) lev=0; else if (lev>=NUM_SCALES) lev = NUM_SCALES-1;
      text("Grain-size = "+nf(scales[lev],3,3),          SZ+MARGIN, 10*LINEHEIGHT);
      break;
    case PARTIALLY_COHERENT_DISARRAY:
      text("UP-DOWN arrow keys to adjust reach",         SZ+MARGIN, 7*LINEHEIGHT);
      text("LEFT-RIGHT arrow keys to adjust coherence",  SZ+MARGIN, 8*LINEHEIGHT);
      text("Reach = "+nf(reach,3,3),                     SZ+MARGIN, 9*LINEHEIGHT);
      text("Coherence = "+nf(coherence,1,3),             SZ+MARGIN, 10*LINEHEIGHT);
      break;
  }
  textFont(font,SMALL_FONT_SIZE);
  fill(BOTTOM_LINE_COLOR);
  text("SPACE_BAR computes new instance",                                   SZ+MARGIN, SZ-2.5f*LINEHEIGHT);
  text("0 fiducial, 1 Lotze, 2 Helmholtz, 3 Coherent, 4 Partial coherence", SZ+MARGIN, SZ-1.5f*LINEHEIGHT);
  text("A disarray instances, B noise fields, C data stacks",               SZ+MARGIN, SZ-0.5f*LINEHEIGHT);
  fill(FONT_COLOR); 
  textFont(font, FONT_SIZE); 
}

public void noiseFieldsMenu(int nf) {
  fill(HEADING_COLOR);
  switch(nf){
    case LOTZE_STACK:
      text("B.0. Lotze disarray",                                                      SZ+MARGIN, LINEHEIGHT);
      fill(FONT_COLOR);
      text("UP/DOWN keys to change grain",                                             SZ+MARGIN,  3*LINEHEIGHT);
      text("Grain = "+nf(effectiveScales[stackLevel],3,3),                             SZ+MARGIN,  4*LINEHEIGHT);
      break;
    case HELMHOLTZ_STACK:
      text("B.1. Helmholtz stack",                                                     SZ+MARGIN, LINEHEIGHT);
      fill(FONT_COLOR);
      text("UP/DOWN keys to change level",                                             SZ+MARGIN,  3*LINEHEIGHT);
      text("Level = "+stackLevel+" (sigma = "+nf(effectiveScales[stackLevel],3,3)+")", SZ+MARGIN,  4*LINEHEIGHT);
      break;
    case COHERENT_STACK:
      text("B.2. Coherent stack",                                                      SZ+MARGIN, LINEHEIGHT);
      fill(FONT_COLOR);
      text("UP/DOWN keys to change level",                                             SZ+MARGIN,  3*LINEHEIGHT);
      text("Level = "+stackLevel+" (sigma = "+nf(effectiveScales[stackLevel],3,3)+")", SZ+MARGIN,  4*LINEHEIGHT);
      break;
  }
  
  textFont(font,SMALL_FONT_SIZE);
  fill(BOTTOM_LINE_COLOR);
  text("SPACE_BAR computes new instance",                                              SZ+MARGIN, SZ-2.5f*LINEHEIGHT);
  text("0 Lotze disarray, 1 Helmholtz stack, 2 Coherent stack",                        SZ+MARGIN, SZ-1.5f*LINEHEIGHT);
  text("A disarray instances, B noise fields, C data stacks",                          SZ+MARGIN, SZ-0.5f*LINEHEIGHT);
  fill(FONT_COLOR); 
  textFont(font, FONT_SIZE); 
}

public void dataStacksMenu(int ds) {
  fill(HEADING_COLOR);
  switch(ds){
    case IMAGE_STACK:
      text("C.0. Image scale space", SZ+MARGIN, LINEHEIGHT);
      break;
    case MACCHIE:
      text("C.1. Macchie scale space", SZ+MARGIN, LINEHEIGHT);
      break;
    case LAPLACEAN_STACK:
      text("C.2. Pointlet stack ('D.O.G.s')", SZ+MARGIN, LINEHEIGHT);
      break;
    case EDGE_FINDER_STACK:
      text("C.3. Edginess stack ('edge finders')", SZ+MARGIN, LINEHEIGHT);
      image(directionColorScale(),1.7f*SZ,0.3f*SZ,0.25f*SZ,0.25f*SZ);
      textFont(font,SMALL_FONT_SIZE);
      fill(255);
      text("direction color scale",1.3f*SZ,0.4f*SZ);
      textFont(font,FONT_SIZE);
      break;
    case EDGELETS:
      text("C.4. Edgelets stack", SZ+MARGIN, LINEHEIGHT);
      break;
    case LINE_FINDER_STACK:
      text("C.5. Ridginess stack ('line finders')", SZ+MARGIN, LINEHEIGHT);
      image(lineFinderBasisColorScale(),1.7f*SZ,0.3f*SZ,0.25f*SZ,0.25f*SZ);
      textFont(font,SMALL_FONT_SIZE);
      fill(255);
      text("orientation color scale",1.3f*SZ,0.4f*SZ);
      textFont(font,FONT_SIZE);
      break;
  }
  fill(FONT_COLOR);
  text("UP/DOWN keys to change level",                                    SZ+MARGIN,  3*LINEHEIGHT);
  text("Level = "+stackLevel+" (sigma = "+nf(scales[stackLevel],3,3)+")", SZ+MARGIN,  4*LINEHEIGHT);
  textFont(font,SMALL_FONT_SIZE);
  fill(BOTTOM_LINE_COLOR);
  text("SPACE_BAR computes new instance",                    SZ+MARGIN, SZ-2.5f*LINEHEIGHT);
  text("0 Image, 1 Macchie, 2 DOGs, 3 1st ord., 4 Edgelets, 5 2nd ord.",    SZ+MARGIN, SZ-1.5f*LINEHEIGHT);
  text("A disarray instances, B noise fields, C data stacks",SZ+MARGIN, SZ-0.5f*LINEHEIGHT);
  fill(FONT_COLOR); 
  textFont(font, FONT_SIZE); 
}

public String[] stateDescription() {
  StringList sl = new StringList();
  switch (mode){
    case DISARRAY_RESULT : {sl.append("Disarray result"); break;}
    case NOISE_FIELDS    : {sl.append("Noise field"); break;}
    case DATA_STACKS     : {sl.append("Datastack"); break;}
  }
  switch (resultDisplayMode){
    case FIDUCIAL_IMAGE              : {sl.append("Fiducial image"); break;}
    case LOTZE_DISARRAY              : {sl.append("Lotze disarray"); break;}
    case HELMHOLTZ_DISARRAY          : {sl.append("Helmholtz disarray"); break;}
    case COHERENT_DISARRAY           : {sl.append("Coherent disarray"); break;}
    case PARTIALLY_COHERENT_DISARRAY : {sl.append("Partially coherent disarray"); break;}
  }
  switch (theNoiseField){
    case LOTZE_STACK      : {sl.append("Lotze stack"); break;}
    case HELMHOLTZ_STACK  : {sl.append("Helmholtz stack"); break;}
    case COHERENT_STACK   : {sl.append("Coherent stack"); break;}
  }
  switch (theStack){
    case IMAGE_STACK       : {sl.append("Image scale space"); break;}
    case MACCHIE           : {sl.append("Macchie");}
    case LAPLACEAN_STACK   : {sl.append("Laplacean scale space"); break;}
    case EDGE_FINDER_STACK : {sl.append("Edginess stack"); break;}
    case EDGELETS          : {sl.append("Edgelets stack"); break;}
    case LINE_FINDER_STACK : {sl.append("Line finder stack"); break;}
  }
  sl.append("scale = "+scales[stackLevel]);
  sl.append("effective scale = "+effectiveScales[stackLevel]);
  sl.append("stackLevel = "+stackLevel);
  sl.append("reach = "+reach);
  sl.append("coherence  = "+coherence);
  return sl.array();
}

public void limitManualPage(int numPages) {
  if      (manualPage > numPages-1) manualPage =          0;
  else if (manualPage <          0) manualPage = numPages-1;
}

public void manual(int mode, int type) {
  background(255);
  if (helpMode) { limitManualPage(HELP.size()); image(HELP.get(manualPage),0,0,2*SZ,SZ);}
  else
    switch (mode) {
      case DISARRAY_RESULT:
        switch (type) {
          case FIDUCIAL_IMAGE:              limitManualPage(A_0.size()); image(A_0.get(manualPage),0,0,2*SZ,SZ); break;
          case LOTZE_DISARRAY:              limitManualPage(A_1.size()); image(A_1.get(manualPage),0,0,2*SZ,SZ); break;
          case HELMHOLTZ_DISARRAY:          limitManualPage(A_2.size()); image(A_2.get(manualPage),0,0,2*SZ,SZ); break;
          case COHERENT_DISARRAY:           limitManualPage(A_3.size()); image(A_3.get(manualPage),0,0,2*SZ,SZ); break;
          case PARTIALLY_COHERENT_DISARRAY: limitManualPage(A_4.size()); image(A_4.get(manualPage),0,0,2*SZ,SZ); break;
        }
        break;
      case NOISE_FIELDS: 
        switch (type) {
          case LOTZE_STACK:                 limitManualPage(B_0.size()); image(B_0.get(manualPage),0,0,2*SZ,SZ); break;
          case HELMHOLTZ_STACK:             limitManualPage(B_1.size()); image(B_1.get(manualPage),0,0,2*SZ,SZ); break;
          case COHERENT_STACK:              limitManualPage(B_2.size()); image(B_2.get(manualPage),0,0,2*SZ,SZ); break;
        }
        break;
      case DATA_STACKS: 
        switch(type){
          case IMAGE_STACK:                 limitManualPage(C_0.size()); image(C_0.get(manualPage),0,0,2*SZ,SZ); break;
          case MACCHIE:                     limitManualPage(C_1.size()); image(C_1.get(manualPage),0,0,2*SZ,SZ); break;
          case LAPLACEAN_STACK:             limitManualPage(C_2.size()); image(C_2.get(manualPage),0,0,2*SZ,SZ); break;
          case EDGE_FINDER_STACK:           limitManualPage(C_3.size()); image(C_3.get(manualPage),0,0,2*SZ,SZ); break;
          case EDGELETS:                    limitManualPage(C_4.size()); image(C_4.get(manualPage),0,0,2*SZ,SZ); break;
          case LINE_FINDER_STACK:           limitManualPage(C_5.size()); image(C_5.get(manualPage),0,0,2*SZ,SZ); break;
        }
        break;
    }
}

public void loadManualPages() {
  // HELP
  HELP.add(loadImage("HELP.png"));
  // A
  A_0.add(loadImage("A0a.png"));
  A_0.add(loadImage("A0b.png"));
  A_0.add(loadImage("A0c.png"));
  
  A_1.add(loadImage("A1a.png"));
  A_1.add(loadImage("A1b.png"));
  A_1.add(loadImage("A1c.png"));
  
  A_2.add(loadImage("A2.png"));
  
  A_3.add(loadImage("A3a.png"));
  A_3.add(loadImage("A3b.png"));
  A_3.add(loadImage("A3c.png"));
  A_3.add(loadImage("A3d.png"));
  A_3.add(loadImage("A3e.png"));
  
  A_4.add(loadImage("A4.png"));
  
  // B
  B_0.add(loadImage("B0.png"));
  
  B_1.add(loadImage("B1.png"));
  
  B_2.add(loadImage("B2.png"));
  
  // C
  C_0.add(loadImage("C0a.png"));
  C_0.add(loadImage("C0b.png"));
  C_0.add(loadImage("C0c.png"));
  
  C_1.add(loadImage("C1a.png"));
  C_1.add(loadImage("C1b.png"));
  
  C_2.add(loadImage("C2.png"));
  
  C_3.add(loadImage("C3a.png"));
  C_3.add(loadImage("C3b.png"));
  
  C_4.add(loadImage("C4a.png"));
  C_4.add(loadImage("C4b.png"));
  C_4.add(loadImage("C4c.png"));
  C_4.add(loadImage("C4d.png"));
  
  C_5.add(loadImage("C5a.png"));
  C_5.add(loadImage("C5b.png"));
  C_5.add(loadImage("C5c.png"));
  C_5.add(loadImage("C5d.png"));
  C_5.add(loadImage("C5e.png"));
} 
static final float LEVEL = 0.25f;

// Generate complex normal deviates
public float[][] randomDataComplex() {
 float[][] data = new float[SZ][2*SZ];
 for (int i=0; i<SZ; i++) for (int j=0; j<2*SZ; j++)  data[i][j] = rnd.nextGaussian();
 return data;
}

// Normalize a data plane
public float[][] zeroMeanUnitVariance(float[][] data) {
  float[] tmp = meanAndStandarddeviation(data);
  float mn = tmp[0];
  float sd = tmp[1];
  float[][] normalized = new float[SZ][SZ];
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++)  normalized[i][j] = (data[i][j]-mn)/sd;
  return normalized;
}

// This generates a random complex spectrum, multiplies it with a blur spectrum, and does an inverse FFT
// The real and imaginary parts are independent zero mean unit variance Gaussian noise fields 
public ArrayList<float[][]> twoRandomFields(int theLevel) {
  // generate complex spectrum - Notice the NORMAL version is slower and you'll hardly notice a difference
  float[][] spectrum = randomDataComplex();
  // Remove DC-component
  spectrum[0][0] = 0.0f;  
  // Multiply with kernel spectrum
  float[][] cs = complexMultiply(spectrum,blurSpectra.get(theLevel));
  // FFT back transform
  fft.complexInverse(cs,true);
  // Split real and imaginary parts
  float[][] re = new float[SZ][SZ];
  float[][] im = new float[SZ][SZ];
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) {
    re[i][j] = cs[i][2*j];
    im[i][j] = cs[i][2*j+1];
  }
  // return two independent noise fields
  ArrayList<float[][]> result = new ArrayList<float[][]>();
  result.add(zeroMeanUnitVariance(re));
  result.add(zeroMeanUnitVariance(im));
  return result;
}

// Random Gaussian noise stack, zero mean, unit variance
public ArrayList<ArrayList<float[][]>> randomGaussianDataStacks() {
  ArrayList<ArrayList<float[][]>> noiseStacks = new ArrayList<ArrayList<float[][]>>();
  ArrayList<float[][]> noiseStackX = new ArrayList<float[][]>();
  ArrayList<float[][]> noiseStackY = new ArrayList<float[][]>();
  for (int k=0; k<NUM_SCALES; k++) {
    ArrayList<float[][]> tmp = twoRandomFields(k);
    noiseStackX.add(tmp.get(0));
    noiseStackY.add(tmp.get(1));
  }
  noiseStacks.add(noiseStackX);
  noiseStacks.add(noiseStackY);
  return noiseStacks;
}

// Random Gaussian noise stack, zero mean, standard deviation sigma
public ArrayList<ArrayList<float[][]>> randomGaussianScaledDataStack() {
  ArrayList<ArrayList<float[][]>> incoherentNoiseStacks = randomGaussianDataStacks();
  for (int k=0; k<NUM_SCALES; k++) for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) {
    incoherentNoiseStacks.get(0).get(k)[i][j] *= effectiveScales[k];
    incoherentNoiseStacks.get(1).get(k)[i][j] *= effectiveScales[k];
  }
  return incoherentNoiseStacks;
}

// Coherent fractal noise stack, zero mean, unit variance
public ArrayList<ArrayList<float[][]>> coherentGaussianDataStack() {
  ArrayList<ArrayList<float[][]>> incoherentNoiseStacks = randomGaussianDataStacks();
  ArrayList<float[][]> coherentNoiseStackX   = new ArrayList<float[][]>();
  ArrayList<float[][]> coherentNoiseStackY   = new ArrayList<float[][]>();
  // BUILD THE STACK IN REVERSE ORDER
  // Put the lowest resolution in the stack
  float[][] tmpX = new float[SZ][SZ];
  float[][] tmpY = new float[SZ][SZ];
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) {
    tmpX[i][j] = incoherentNoiseStacks.get(0).get(NUM_SCALES-1)[i][j];
    tmpY[i][j] = incoherentNoiseStacks.get(0).get(NUM_SCALES-1)[i][j];
  }
  coherentNoiseStackX.add(tmpX);
  coherentNoiseStackY.add(tmpY);
  // For each finer layer put the sum of the coarse fractal and the noise layer on the stack
  for (int k=NUM_SCALES-1; k>=0; k--) {
    int n = coherentNoiseStackX.size() - 1;
    float[][] tmp2X = new float[SZ][SZ];
    float[][] tmp2Y = new float[SZ][SZ];
    for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) {
      tmp2X[i][j] = incoherentNoiseStacks.get(0).get(k)[i][j] + coherentNoiseStackX.get(n)[i][j];
      tmp2Y[i][j] = incoherentNoiseStacks.get(1).get(k)[i][j] + coherentNoiseStackY.get(n)[i][j];
    }
    coherentNoiseStackX.add(tmp2X);
    coherentNoiseStackY.add(tmp2Y);
  }
  // PUT STACK UPSIDE DOWN (==RIGHT AGAIN!)
  ArrayList<float[][]> coherentNoiseStack2X = new ArrayList<float[][]>();
  ArrayList<float[][]> coherentNoiseStack2Y = new ArrayList<float[][]>();
  for (int k=NUM_SCALES-1; k>=0; k--) {
    coherentNoiseStack2X.add(coherentNoiseStackX.get(k));
    coherentNoiseStack2Y.add(coherentNoiseStackY.get(k));
  }
  ArrayList<ArrayList<float[][]>> coherentNoiseStacks2 = new ArrayList<ArrayList<float[][]>>();
  coherentNoiseStacks2.add(coherentNoiseStack2X);
  coherentNoiseStacks2.add(coherentNoiseStack2Y);
  return coherentNoiseStacks2;
}

// Estimate the "amplitude" of a stack
public float stacksAmplitude(ArrayList<ArrayList<float[][]>> stacks, float level){
  FloatList d = new FloatList();
  for (int k=0; k<NUM_SCALES; k++) 
    for (int i=0; i<SZ; i++) 
      for (int j=0; j<SZ; j++) 
        d.append(sqrt(sq(stacks.get(0).get(k)[i][j]) + sq(stacks.get(1).get(k)[i][j])));
  d.sort();
  // return the quantile as a measure of "stack amplitude"
  return d.get(round(level*SZ*SZ));
}

// Normalise X, Y stacks to unit amplitude
public ArrayList<ArrayList<float[][]>> normaliseStacks(ArrayList<ArrayList<float[][]>> stacks, float level) {
  float a = stacksAmplitude(stacks,level);
  ArrayList<ArrayList<float[][]>> stacksN = new ArrayList<ArrayList<float[][]>>();
  ArrayList<float[][]> stackX = new ArrayList<float[][]>();
  ArrayList<float[][]> stackY = new ArrayList<float[][]>();
  for (int k=0; k<NUM_SCALES; k++) {
    float[][] x = new float[SZ][SZ];
    float[][] y = new float[SZ][SZ];
    for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) {
      x[i][j] = stacks.get(0).get(k)[i][j]/a;
      y[i][j] = stacks.get(1).get(k)[i][j]/a;
    }
    stackX.add(x);
    stackY.add(y);
  }
  stacksN.add(stackX);
  stacksN.add(stackY);
  return stacksN;
}

public void computeRandomDisarrayScaleSpaces() {
  initializationReport = initializationReport+"\n"+"Computing random disarray stacks";
  redraw();
  int epoch = millis();
  computeRandomDisarrayStacks();
  computeRandomScaledDisarrayStacks();
  computeCoherentDisarrayStacks();
  initializationReport = initializationReport+"\n"+"Computed random disarray stacks in "+(millis()-epoch)+" ms";
  redraw();
}

public void computeRandomDisarrayStacks() {
  ArrayList<ArrayList<float[][]>> tmp = randomGaussianDataStacks();
  ArrayList<ArrayList<float[][]>> tmp2 = normaliseStacks(tmp,LEVEL);
  randomGaussianDataStackX.clear();
  randomGaussianDataStackY.clear();
  randomGaussianDataStackX = tmp2.get(0);
  randomGaussianDataStackY = tmp2.get(1);
}

public void computeRandomScaledDisarrayStacks() {
  ArrayList<ArrayList<float[][]>> tmp = randomGaussianScaledDataStack();
  ArrayList<ArrayList<float[][]>> tmp2 = normaliseStacks(tmp,LEVEL);
  randomGaussianScaledDataStackX.clear();
  randomGaussianScaledDataStackY.clear();
  randomGaussianScaledDataStackX = tmp2.get(0);
  randomGaussianScaledDataStackY = tmp2.get(1);
}

public void computeCoherentDisarrayStacks() {
  ArrayList<ArrayList<float[][]>> tmp = coherentGaussianDataStack();
  ArrayList<ArrayList<float[][]>> tmp2 = normaliseStacks(tmp,LEVEL);
  coherentDataStackX.clear();
  coherentDataStackY.clear();
  coherentDataStackX = tmp2.get(0);
  coherentDataStackY = tmp2.get(1);
}
/* ************************************************************************ */
/*                                                                          */
/*                         GENERATION OF RANDOM NUMBERS                     */   
/*                         USES A QUALITY PCG ALGORITHM                     */
/*                                                                          */
/* ************************************************************************ */



class Pcg32 {
  long state;
  long inc;

  final static long MULTIPLIER = 6364136223846793005L;

  Pcg32(long initState, long initSeq) {
      seed(initState, initSeq);
  }

  Pcg32() { seed();}

  // Returns the next pseudorandom, approximately uniformly distributed int
  // value from this random number generator's sequence. 
  public int nextInt() {
      long oldState = state;
      state = oldState * MULTIPLIER + inc;
      int xorShifted = (int) (((oldState >>> 18) ^ oldState) >>> 27);
      int rot = (int) (oldState >>> 59);
      return Integer.rotateRight(xorShifted, rot);
  }
   
  // Returns the next pseudorandom, approximately uniformly distributed int
  // value between 0 (inclusive) and {@code n} (exclusive).
  public int nextInt(int n) {
      if (n <= 0) {
          throw new IllegalArgumentException("n must be positive");
      }
      // the special treatment of powers of 2 as in Random.nextInt(int) shouldn't be necessary
      int bits, val;
      do {
          bits = nextInt() >>> 1;
          val = bits % n;
      } while (bits - val + (n-1) < 0);
      return val;
  }

   // Returns the next pseudorandom, approximately uniformly distributed boolean value.
  public boolean nextBoolean() { return (nextInt() & 1) != 0;}

   // Returns the next pseudorandom, approximately uniformly distributed float value between 0.0 (inclusive)
   // and 1.0 (exclusive)
 public float nextFloat() { return nextBits(24) / ((float)(1 << 24));}

   // Returns the next pseudorandom, approximately uniformly distributed float value between 0.0 (inclusive)
   // and bound (exclusive).
  public float nextFloat(float bound) { return bound * nextFloat(); }
  
      public int nextBits(int bits) {return nextInt() >>> (32 - bits);}
  
  float nextNextGaussian;
  boolean haveNextNextGaussian = false;

   // Returns the next pseudorandom, Gaussian ("normally") distributed
   // float value with mean 0.0 and standard
   // deviation 1.0 from this random number generator's sequence.
  public float nextGaussian() {
      // See Knuth, ACP, Section 3.4.1 Algorithm C.
      if (haveNextNextGaussian) {
          haveNextNextGaussian = false;
          return nextNextGaussian;
      } else {
          float v1, v2, s;
          do {
              v1 = 2 * nextFloat() - 1; // between -1 and 1
              v2 = 2 * nextFloat() - 1; // between -1 and 1
              s = v1 * v1 + v2 * v2;
          } while (s >= 1 || s == 0);
          float multiplier = sqrt(-2.0f * log(s)/s);
          nextNextGaussian = v2 * multiplier;
          haveNextNextGaussian = true;
          return v1 * multiplier;
      }
  }

   // Returns the next pseudorandom, Gaussian ("normally") distributed
   // double value with given mean and standard
   // deviation from this random number generator's sequence.
  public float nextGaussian(float mean, float standardDeviation) { return nextGaussian() * standardDeviation + mean;}

  public void seed(long initState, long initSeq) {
      state = 0;
      inc = 2 * initSeq + 1;
      nextInt();
      state += initState;
      nextInt();
  }

   // Initializes the generator with current time as the state and
   // a (very likely to be) unique stream number. This ensures that
   // even if multiple generators are initialized using this method
   // while System#nanoTime() returns the same time, they
   // will have different sequences.
   
  // JUST A SANITY CHECK
   
  public void seed() { seed(System.nanoTime(), streamUniquifier());}

  public long streamUniquifier() {
      for (;;) {
          long current = streamUniquifier.get();
          long next = current * 181783497276652981L;
          if (streamUniquifier.compareAndSet(current, next)) return next;
      }
  }

  final AtomicLong streamUniquifier = new AtomicLong(System.identityHashCode(Pcg32.class));
}
float[] scales          = new float[NUM_SCALES];
float[] effectiveScales = new float[NUM_SCALES];
ArrayList<float[][]> blurSpectra = new ArrayList<float[][]>(); 

PImage               fiducialImage;                                            
float[][]            fiducialData;                                             
float[][]            fiducialSpectrum;  

ArrayList<float[][]> fiducialScaleSpace          = new ArrayList<float[][]>();
ArrayList<float[][]> fiducialScaleLayers         = new ArrayList<float[][]>();
ArrayList<float[][]> fiducialFirstOrderStackX    = new ArrayList<float[][]>();
ArrayList<float[][]> fiducialFirstOrderStackY    = new ArrayList<float[][]>();
ArrayList<float[][]> fiducialSecondOrderStackP   = new ArrayList<float[][]>();
ArrayList<float[][]> fiducialSecondOrderStackQ   = new ArrayList<float[][]>(); 
ArrayList<float[][]> fiducialSecondOrderStackR   = new ArrayList<float[][]>();

ArrayList<float[][]> randomGaussianDataStackX        = new ArrayList<float[][]>();
ArrayList<float[][]> randomGaussianDataStackY        = new ArrayList<float[][]>();
ArrayList<float[][]> randomGaussianScaledDataStackX  = new ArrayList<float[][]>();
ArrayList<float[][]> randomGaussianScaledDataStackY  = new ArrayList<float[][]>();
ArrayList<float[][]> coherentDataStackX              = new ArrayList<float[][]>(); 
ArrayList<float[][]> coherentDataStackY              = new ArrayList<float[][]>();

String initializationReport = "Initialization stage\n\nStarting with scale space";

// Set up everything connected with scales and the fiducial image
public void getImage(String path) {
  int epoch = millis();
  // Try the image for size
  fiducialImage = loadImage(path);
  fiducialImage.resize(SZ,SZ); // Force standard size
  // Define scales
  scales[0] = 1.0f/sqrt(2.0f);
  for (int k=1; k<NUM_SCALES; k++) scales[k] = scales[k-1]*sqrt(2.0f);
  for (int k=0; k<NUM_SCALES-1; k++) effectiveScales[k] = sqrt(scales[k]*scales[k+1]);
  effectiveScales[NUM_SCALES-1] =  effectiveScales[NUM_SCALES-2]*sqrt(2.0f);
  initializationReport = initializationReport+"\n"+"Effective scales run from "+nf(effectiveScales[0],3,2)+" to "+nf(effectiveScales[NUM_SCALES-1],3,0);
  redraw();
  // store the image as float buffer and compute its complex spectrum
  fiducialData     = imageToData(fiducialImage);
  fiducialSpectrum = imageSpectrum(fiducialData);
  // compute the image scale space
  fiducialScaleSpace.clear();
  for (int i=0; i<NUM_SCALES; i++) fiducialScaleSpace.add(runConvolution(fiducialSpectrum,kernelSpectrum(scales[i])));
  // compute the image scale layer increments
  fiducialScaleLayers.clear();
  for (int k=0; k<NUM_SCALES-1; k++) {
    float[][] diff = new float[SZ][SZ];
    for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) diff[i][j] = fiducialScaleSpace.get(k)[i][j]-fiducialScaleSpace.get(k+1)[i][j];
    fiducialScaleLayers.add(diff);
  }
  float[][] diff = new float[SZ][SZ];
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) diff[i][j] = fiducialScaleSpace.get(NUM_SCALES-1)[i][j];
  fiducialScaleLayers.add(diff);
  // Compute complex blur spectra - this allows fast blurring when needed (AT LEAST IF USED!)
  blurSpectra.clear();
  for (int i=0; i<NUM_SCALES; i++) blurSpectra.add(kernelSpectrum(effectiveScales[i]));
  initializationReport = initializationReport+"\n"+"Set up took "+(millis()-epoch)+" ms";
  redraw();
}

public void computeFirstOrderStack() {
  int epoch = millis();
  fiducialFirstOrderStackX.clear();
  fiducialFirstOrderStackY.clear();
  for (int k=0; k<NUM_SCALES; k++) { // Here I compute a Cartesian tangent vector basis
    float[][] x, y;
    fiducialFirstOrderStackX.add(runConvolution(fiducialSpectrum, kernelSpectrum(1,0,effectiveScales[k])));
    fiducialFirstOrderStackY.add(runConvolution(fiducialSpectrum, kernelSpectrum(0,1,effectiveScales[k])));
  }
  initializationReport = initializationReport+"\n"+"Computed first order in "+(millis()-epoch)+" ms";
  redraw();
}

public void computeCorticalRepresentation(String path) {
  int epoch = millis();
  getImage(path); 
  computeRandomDisarrayScaleSpaces();
  initializationReport = initializationReport+"\n"+"Computing 1st order stack";
  redraw();
  computeFirstOrderStack();
  initializationReport = initializationReport+"\n"+"Computing 2nd order stack";
  redraw();
  computeSecondOrderStack();
  initializationReport = initializationReport+"\n"+"Computed cortical representation up to 2nd order in "+(millis()-epoch)+" ms";
  redraw();
  imageHasBeenDefined = true;
  firsttime = true;
}

public void computeSecondOrderStack() {
  int epoch = millis();
  // Find the Hessiam matrix
  ArrayList<float[][]> hxx = new ArrayList<float[][]>();
  ArrayList<float[][]> hxy = new ArrayList<float[][]>();
  ArrayList<float[][]> hyy = new ArrayList<float[][]>();
  for (int k=0; k<NUM_SCALES; k++) { // Here I compute the coefficients of the Hessian matrix
    float[][] x, y;
    hxx.add(runConvolution(fiducialSpectrum, kernelSpectrum(2,0,effectiveScales[k])));
    hxy.add(runConvolution(fiducialSpectrum, kernelSpectrum(1,1,effectiveScales[k])));
    hyy.add(runConvolution(fiducialSpectrum, kernelSpectrum(0,2,effectiveScales[k])));
  }
  // Express in terms of line finder basis, directions n Pi/3, n = 0, 1, 2
  float z = 0.5f*sqrt(3);
  fiducialSecondOrderStackP.clear();
  fiducialSecondOrderStackQ.clear();
  fiducialSecondOrderStackR.clear();
  for (int k=0; k<NUM_SCALES; k++) {
    float[][] tmp0 = new float[SZ][SZ];
    float[][] tmp1 = new float[SZ][SZ];
    float[][] tmp2 = new float[SZ][SZ];
    for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) {
      float zz = 0.25f*hxx.get(k)[i][j]+0.75f*hyy.get(k)[i][j]; // Here I transform from the Hessian matrix representation
      tmp0[i][j] =      hxx.get(k)[i][j];                     // To a nice, isotropic basis of three line finders at 60\u00ba orientation differences
      tmp1[i][j] = zz-z*hxy.get(k)[i][j];                     // The coefficients are easy enough to find with a little algebra
      tmp2[i][j] = zz+z*hxy.get(k)[i][j];
    }
    fiducialSecondOrderStackP.add(tmp0);
    fiducialSecondOrderStackQ.add(tmp1);
    fiducialSecondOrderStackR.add(tmp2);
  }
  initializationReport = initializationReport+"\n"+"Computed second order in "+(millis()-epoch)+" ms";
  redraw();
}

public float[][] firstOrderDirectionality(float phi, int level) {
  float[][] result = new float[SZ][SZ];
  float c = cos(radians(phi));
  float s = sin(radians(phi));
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) 
    result[i][j] =c*fiducialFirstOrderStackX.get(level)[i][j]+s*fiducialFirstOrderStackY.get(level)[i][j];
  return result;
}

public float[][] secondOrderDirectionality(float phi, int level) {
  float[][] result = new float[SZ][SZ];
  float u = (1.0f+2.0f*cos(2.0f*radians(phi)))/3.0f;                 // This is easy enough to prove with a little algebra
  float v =  4.0f*cos(PI/6.0f-radians(phi))*sin(radians(phi))/3.0f;
  float w = -4.0f*cos(PI/6.0f+radians(phi))*sin(radians(phi))/3.0f;
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) 
    result[i][j] = u*fiducialSecondOrderStackP.get(level)[i][j]+v*fiducialSecondOrderStackQ.get(level)[i][j]+w*fiducialSecondOrderStackR.get(level)[i][j];
  return result;
}
  public void settings() {  size(1024,512,JAVA2D); }
  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "EidolonFactory" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
