String PATH = "Face.jpg"; // Default image
boolean imageHasBeenDefined = false;
boolean userStartedInterface = false;
boolean firsttime = false;

static final int SZ         = 512;
static final int NUM_SCALES = 14;

FloatFFT_2D fft;
Pcg32       rnd;
int         initialEpoch;

void setup() {
  size(1024,512,JAVA2D); 
  selectInput("Select a file to process:", "fileSelected");
  initialEpoch = millis();
  fft = new FloatFFT_2D(SZ,SZ);
  rnd = new Pcg32(); 
  font = loadFont("ComicSansMS-32.vlw");
  altFont = loadFont("Stencil-48.vlw");
  loadManualPages();
  // computeCorticalRepresentation(PATH);
  textFont(font, FONT_SIZE); 
  fill(FONT_COLOR); 
  noStroke(); 
  noLoop();
  surface.setTitle("Eidolon Factory");
}

void fileSelected(File selection) {
  if (selection == null) {
    initializationReport = initializationReport+"\n"+"Computations for default image";
    redraw();
    computeCorticalRepresentation(PATH); // Use default image
    redraw();
  } 
  else {
    initializationReport = initializationReport+"\n"+"Computations for user selected image";
    computeCorticalRepresentation(selection.getAbsolutePath()); // User user selected image
    redraw();
  }
}

void draw() {
  background(0);
  if (imageHasBeenDefined && userStartedInterface) {
    if (helpMode) {manual(-1,-1); helpMode=false;}
    else
      switch(mode) {
        case DISARRAY_RESULT: 
          if (displayMode==DEMO_MODE) {
            displayResults(resultDisplayMode); 
            resultsMenu(resultDisplayMode); 
          } 
          else manual(DISARRAY_RESULT,resultDisplayMode);
          break;
        case NOISE_FIELDS:
          if (displayMode==DEMO_MODE) {
            displayNoiseFields(theNoiseField);
            noiseFieldsMenu(theNoiseField); 
          } 
          else manual(NOISE_FIELDS,theNoiseField);
          break;
        case DATA_STACKS:
          if (displayMode==DEMO_MODE) {
            displayDataStack(theStack);
            dataStacksMenu(theStack); 
          } 
          else manual(DATA_STACKS,theStack);
          break;
      }
  }
  else {
    image(loadImage("Frontispiece.png"),0,0,2*SZ,SZ);
    fill(0,255,0);
    textFont(font,SMALL_FONT_SIZE);
    text(initializationReport,650,50);
    if (imageHasBeenDefined) {
      fill(255,255,0);
      textFont(font,SMALL_FONT_SIZE);
      text("ready when you are!",650,400);
      textFont(altFont,48);
      text("CLICKME!",650,450);
    }
    textFont(font,SMALL_FONT_SIZE);
  }
}

void displayResults(int rdm) {
  int epoch = millis();
  fill(HEADING_COLOR);
  switch (rdm) {
    case FIDUCIAL_IMAGE:
      int lev = round(coherence*NUM_SCALES);
      if (lev<0) lev=0; else if (lev>=NUM_SCALES) lev = NUM_SCALES-1;
      image(captcha(fiducialImage, lev, reach),0,0,SZ,SZ);
      if (firsttime) {  
        fill(255,0,0);
        textFont(altFont,48);
        text("CLICKME!",100,100);
        firsttime = false;
      }
      fill(255,255,0);
      textFont(font,SMALL_FONT_SIZE);
      text("A.0. CAPTCHA of fiducial image",                                SZ+MARGIN, LINEHEIGHT);
      // text("Eidolon factory set up time "+(millis()-initialEpoch)+" ms", SZ+MARGIN, 2*LINEHEIGHT);
      break;
    case LOTZE_DISARRAY:
      lev = round(coherence*NUM_SCALES);
      if (lev<0) lev=0; else if (lev>=NUM_SCALES) lev = NUM_SCALES-1;
      computeRandomDisarrayStacks();
      theCurrentlyDisplayedImage = showLocalDisarrayLotze(lev,reach);
      image(theCurrentlyDisplayedImage,0,0,SZ,SZ);
      text("A.1. Lotze-type disarray ("+(millis()-epoch)+" ms)",              SZ+MARGIN, LINEHEIGHT);
      break;
   case HELMHOLTZ_DISARRAY:
      computeRandomScaledDisarrayStacks();
      theCurrentlyDisplayedImage = showLocalDisarrayHelmholtz(reach);
      image(theCurrentlyDisplayedImage,0,0,SZ,SZ);
      text("A.2. Helmholtz-type disarray ("+(millis()-epoch)+" ms)",           SZ+MARGIN, LINEHEIGHT);
      break;
    case COHERENT_DISARRAY:
      computeCoherentDisarrayStacks();
      computeCoherentDisarrayStacks();
      theCurrentlyDisplayedImage = showLocalDisarrayCoherent(reach);
      image(theCurrentlyDisplayedImage,0,0,SZ,SZ);
      text("A.3. Coherent disarray ("+(millis()-epoch)+" ms)",                 SZ+MARGIN, LINEHEIGHT);
      break;
    case PARTIALLY_COHERENT_DISARRAY:
      computeRandomScaledDisarrayStacks();
      computeCoherentDisarrayStacks();
      computeRandomScaledDisarrayStacks();
      theCurrentlyDisplayedImage = showLocalDisarrayPartiallyCoherent(reach,coherence);
      image(theCurrentlyDisplayedImage,0,0,SZ,SZ);
      text("A.4. Partially coherent disarray ("+(millis()-epoch)+" ms)",       SZ+MARGIN, LINEHEIGHT);
      break;
  }
  fill(255,0,0);
  text(fileSaveReport,50,50);
  fileSaveReport = "";
  fill(FONT_COLOR);
}

void displayNoiseFields(int nf) {
  switch(nf){
    case LOTZE_STACK:
      theCurrentlyDisplayedImage = vectorImage(randomGaussianDataStackX.get(stackLevel),randomGaussianDataStackY.get(stackLevel));
      image(theCurrentlyDisplayedImage,0,0,SZ,SZ);
      image(noiseBinCounts(randomGaussianDataStackX.get(stackLevel),
                           randomGaussianDataStackY.get(stackLevel)),1.4*SZ,0.5*SZ,0.5*SZ,0.3*SZ);
      break;
    case HELMHOLTZ_STACK:
      theCurrentlyDisplayedImage = vectorImage(randomGaussianScaledDataStackX.get(stackLevel),randomGaussianScaledDataStackY.get(stackLevel));
      image(theCurrentlyDisplayedImage,0,0,SZ,SZ);
      image(noiseBinCounts(randomGaussianScaledDataStackX.get(stackLevel), 
                           randomGaussianScaledDataStackY.get(stackLevel)),1.4*SZ,0.5*SZ,0.5*SZ,0.3*SZ);
      break;
    case COHERENT_STACK:
      theCurrentlyDisplayedImage = vectorImage(coherentDataStackX.get(stackLevel),coherentDataStackY.get(stackLevel));
      image(theCurrentlyDisplayedImage,0,0,SZ,SZ);
      image(noiseBinCounts(coherentDataStackX.get(stackLevel),             
                           coherentDataStackY.get(stackLevel)),1.4*SZ,0.5*SZ,0.5*SZ,0.3*SZ);
      break;
  }
  textFont(font,SMALL_FONT_SIZE);
  text("Median "+nf(noiseQuantiles[0],3,3),SZ+MARGIN,0.5*SZ-LINEHEIGHT);
  text("75% quantile "+nf(noiseQuantiles[1],3,3),SZ+MARGIN,0.5*SZ-2*LINEHEIGHT);
  text("95% quantile "+nf(noiseQuantiles[2],3,3),SZ+MARGIN,0.5*SZ-3*LINEHEIGHT);
  fill(255);
  text("reach histogram",1.1*SZ,0.6*SZ);
  textFont(font,FONT_SIZE);
}

void displayDataStack(int ds) {
  switch(ds){
    case IMAGE_STACK:
      theCurrentlyDisplayedImage = dataToImage(fiducialScaleSpace.get(stackLevel),0.95);
      image(theCurrentlyDisplayedImage,0,0,SZ,SZ);
      break;
    case MACCHIE:
      theCurrentlyDisplayedImage = macchie(stackLevel);
      image(theCurrentlyDisplayedImage,0,0,SZ,SZ);
      break;
    case LAPLACEAN_STACK:
      theCurrentlyDisplayedImage = dataToImage(fiducialScaleLayers.get(stackLevel),0.95);
      image(theCurrentlyDisplayedImage,0,0,SZ,SZ);
      break;
    case EDGE_FINDER_STACK:
      theCurrentlyDisplayedImage = vectorImage(fiducialFirstOrderStackX.get(stackLevel),fiducialFirstOrderStackY.get(stackLevel));
      image(theCurrentlyDisplayedImage,0,0,SZ,SZ);
      break;
    case EDGELETS:
      theCurrentlyDisplayedImage = edgelets(stackLevel);
      image(theCurrentlyDisplayedImage,0,0,SZ,SZ);
      break;
    case LINE_FINDER_STACK:
      theCurrentlyDisplayedImage = lineFinderBasisImage(fiducialSecondOrderStackP.get(stackLevel),fiducialSecondOrderStackQ.get(stackLevel),fiducialSecondOrderStackR.get(stackLevel));
      image(theCurrentlyDisplayedImage,0,0,SZ,SZ);
      break;
  }
}