import java.util.Arrays;

static final int NOISE_NUM_BINS = 100;
static final float NOISE_CUTOFF = 0.99;
float[] noiseQuantiles = new float[3];

// Extract the green channel of an RGB image as a single float data plane
float[][] imageToData(PImage im) {
 float[][] data = new float[SZ][SZ];
 im.loadPixels();
 for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) data[i][j] = green(im.pixels[j*SZ+i]);
 return data;
}

// Put the data in all color channels of an RGB image
PImage dataToImage(float[][] data) {
 PImage im = new PImage(SZ,SZ);
 im.loadPixels();
 for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) im.pixels[j*SZ+i] = color(data[i][j],data[i][j],data[i][j]);
 im.updatePixels();
 return im;
}

float[][] invertData(float[][] data) {
  float[][] invData = new float[SZ][SZ];
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) invData[i][j] = -data[i][j];
  return invData;
}

// Given float data create a monochrome image
// level is a quantile limit, useful to set the scale
// For level try 0.99, 0.95, or 0.9 say
PImage dataToImage(float[][] buffer, float level){
  float p = quantile(buffer,level);
  float q = quantile(buffer,1-level);
  // Create an image
  PImage im = new PImage(SZ,SZ);
  // Fill a monochrome image using levels between 1-level and level:
  im.loadPixels();
  float tmp = 255.0/(q-p);
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) {
      float val = constrain(tmp*(buffer[i][j]-p),0.0,255.0);
      im.pixels[j*SZ+i] = color(255-val);
    }
  im.updatePixels();
  return im;
}

// Given float data create a monochrome image
// level are the quartiles and 5% limits, useful to set the scale
PImage dataToPoster(float[][] data){
  float[] p = quartilesAnd5PercentLevels(data);
  // Create an image
  PImage im = new PImage(SZ,SZ);
  // Fill a monochrome image using all available levels
  im.loadPixels();
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) 
    if      (data[i][j]<p[0]) im.pixels[j*SZ+i] = color(  0);
    else if (data[i][j]<p[1]) im.pixels[j*SZ+i] = color( 63);
    else if (data[i][j]<p[2]) im.pixels[j*SZ+i] = color(127); 
    else if (data[i][j]<p[3]) im.pixels[j*SZ+i] = color(191);
    else                      im.pixels[j*SZ+i] = color(255);
  im.updatePixels();
  return im;
}

PImage macchie(int theLevel) {
  float[][] data = fiducialScaleSpace.get(theLevel);
  float[] tmp = flattenData(data);
  Arrays.sort(tmp);
  float darkToShade      = tmp[round(    tmp.length/6.0)];
  float shadeToNeutral   = tmp[round(2.0*tmp.length/6.0)];
  float neutralToLight   = tmp[round(4.0*tmp.length/6.0)];
  float lightToHighlight = tmp[round(5.0*tmp.length/6.0)];
  PImage posterized = new PImage(SZ,SZ);
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) {
    float val = data[i][j];
    color c = color(127);
    if      (val<darkToShade)                              c = color(0);
    else if ((val<shadeToNeutral) && (val>darkToShade))    c = color(63);
    else if ((val<neutralToLight) && (val>shadeToNeutral)) c = color(196);
    else if (val>lightToHighlight)                         c = color(255);
    posterized.set(i,j,c);
  }
  return posterized;
}

PImage edgelets(int theLevel) {
  float[][] data = fiducialScaleLayers.get(theLevel);
  float[] tmp = flattenData(data);
  Arrays.sort(tmp);
  float darkToShade      = tmp[round(0.1*tmp.length)];
  float shadeToNeutral   = tmp[round(0.2*tmp.length)];
  float neutralToLight   = tmp[round(0.8*tmp.length)];
  float lightToHighlight = tmp[round(0.9*tmp.length)];
  PImage posterized = new PImage(SZ,SZ);
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) {
    float val = data[i][j];
    color c = color(127);
    if      (val<darkToShade)                              c = color(0);
    else if ((val<shadeToNeutral) && (val>darkToShade))    c = color(63);
    else if ((val<neutralToLight) && (val>shadeToNeutral)) c = color(196);
    else if (val>lightToHighlight)                         c = color(255);
    posterized.set(i,j,c);
  }
  return posterized;
}

float[] flattenData(float[][] data) {
  float[] flat = new float[SZ*SZ];
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) flat[j*SZ+i] = data[i][j];
  return flat;
}

float[] meanAndStandarddeviation(float[][] data) {
  float[] tmp = flattenData(data);
  float s  = 0.0;
  float ss = 0.0;
  for (int i=0; i<SZ*SZ; i++) {
    s  +=    tmp[i];
    ss += sq(tmp[i]);
  }
  // Find mean and standard deviations
  s  /= (SZ*SZ);
  ss /= (SZ*SZ);
  ss = sqrt(abs(sq(s)-ss));
  float[] result = {s,ss};
  return result;
}

int[] binCounts(float[][] data, float mn, float mx, int numBins) {
  float[] tmp = flattenData(data);
  float rn = mx- mn;
  int[] counts = new int[numBins];
  int j;
  for (int i=0;i<SZ*SZ; i++) {
    j = floor(numBins*(tmp[i]-mn)/rn);
    if ((j>=0)&&(j<numBins)) counts[j]++;
  }
  return counts;
}

PImage imageHistogram(float[][] data, int numBins) {
  int[] counts = binCounts(data, 0.0, 255.0, numBins);
  float[] fc = new float[numBins];
  float tmp = 0.0;
  for (int i=0; i<numBins; i++) if (counts[i]>tmp) tmp=counts[i];
  for (int i=0; i<numBins; i++) fc[i] = counts[i]/tmp;
  PGraphics g = createGraphics(SZ,SZ,JAVA2D);
  g.beginDraw();
    g.background(127);
    g.stroke(0);
    g.strokeWeight(0.5);
    g.fill(127,127,255);
    g.rect(50,50,SZ-100,SZ-100);
    g.noStroke();
    float  w = (SZ - 100.0)/numBins;
    for (int i=0; i<numBins; i++) {
      g.fill(255.0*(i+0.5)/numBins);
      g.rect(50+i*w,SZ-50-fc[i]*(SZ-150),w,fc[i]*(SZ-150));
    }
  g.endDraw();
  return g;
}

PImage noiseBinCounts(float[][] layerX, float[][] layerY){
  FloatList d = new FloatList();
  for (int k=0; k<NUM_SCALES; k++) 
    for (int i=0; i<SZ; i++) 
      for (int j=0; j<SZ; j++) 
        d.append(sqrt(sq(layerX[i][j]) + sq(layerY[i][j])));
  d.sort();
  float cutoff = d.get(round(NOISE_CUTOFF*d.size()));
  noiseQuantiles[0] = d.get(round(0.50*d.size()));
  noiseQuantiles[1] = d.get(round(0.75*d.size()));
  noiseQuantiles[2] = d.get(round(0.95*d.size()));
  int numBins = 100;
  int[] counts = new int[numBins];
  int j;
  for (int i=0;i<d.size(); i++) {
    j = floor(NOISE_NUM_BINS*(d.get(i)/cutoff));
    if ((j>=0)&&(j<NOISE_NUM_BINS)) counts[j]++;
  }
  int top = 0;
  for (int i=0;i<NOISE_NUM_BINS; i++) if (counts[i]>top) top=counts[i];
  float[] tmp = new float[NOISE_NUM_BINS];
  for (int i=0;i<NOISE_NUM_BINS; i++) tmp[i] = float(counts[i])/top;
  PGraphics g = createGraphics(SZ,SZ,JAVA2D);
  g.beginDraw();
    g.background(255);
    g.noStroke();
    g.fill(0);
    for (int i=0;i<NOISE_NUM_BINS; i++) g.rect((i*SZ/NOISE_NUM_BINS),(1-tmp[i])*SZ,SZ/NOISE_NUM_BINS,tmp[i]*SZ);
    g.noFill();
    g.stroke(255,255,0);
    g.strokeWeight(8.0);
    g.rect(0,0,SZ,SZ);
  g.endDraw();
  return g;
}

// Return the quartiles of data
float[] quartilesAnd5PercentLevels(float[][] data) {
  float[] tmp = flattenData(data);
  Arrays.sort(tmp);
  float[] q = {tmp[round(0.05*tmp.length)],tmp[round(0.25*tmp.length)],tmp[round(0.5*tmp.length)],tmp[round(0.75*tmp.length)],tmp[round(0.95*tmp.length)],};
  return q;
}

// Return a quantile of data
float quantile(float[][] data, float theLevel) {
  float[] tmp = flattenData(data);
  Arrays.sort(tmp);
  return(tmp[round(theLevel*tmp.length)]);
}

// Planar vector image - hue drom direction, saturation signifies magnitude
PImage vectorImage(float[][] x, float[][] y) {
  float[][] rho = new float[SZ][SZ];
  float[][] phi = new float[SZ][SZ];
  float maxRho = 0.0;
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++)  {
    rho[i][j] = sq(x[i][j])+sq(y[i][j]);
    if (rho[i][j]>maxRho) maxRho = rho[i][j];
    phi[i][j] = atan2(y[i][j],x[i][j]);
  }
  float limit = 0.9; // the maximum saturation
  float cnst = maxRho*(1.0-limit)/limit;
  PImage im = new PImage(SZ,SZ);
  im.loadPixels();
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) 
    im.pixels[j*SZ+i] = lerpColor(color(255,255,255),directionColor(phi[i][j]),rho[i][j]/(cnst+rho[i][j]));
  im.updatePixels();
  return im;
}

// Shows the 3D line finder vector basis in RGB
PImage lineFinderBasisImage(float[][] secondOrderP,float[][] secondOrderQ, float[][] secondOrderR) {
  PImage im = new PImage(SZ,SZ);
  float factor = 127.0/max(quantile(secondOrderP,0.75), quantile(secondOrderQ,0.75), quantile(secondOrderR,0.75));
  im.loadPixels();
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) 
    im.pixels[j*SZ+i] = color(127+factor*secondOrderP[i][j],127+factor*secondOrderQ[i][j],127+factor*secondOrderR[i][j]);
  im.updatePixels();
  return im;
}

PImage lineFinderBasisColorScale() {
  PGraphics g = createGraphics(SZ,SZ,JAVA2D);
  g.beginDraw();
    g.background(0);
    g.ellipseMode(RADIUS);
    g.noStroke();
    float semiSector = PI/6.0;
    float phi0 = 0.0;
    float phi1 = PI/3.0;
    float phi2 = 2.0*PI/3.0;
    g.fill(255,0,0);
    g.arc(0.5*SZ,0.5*SZ,0.5*SZ,0.5*SZ,phi0-semiSector,phi0+semiSector,PIE);
    g.arc(0.5*SZ,0.5*SZ,0.5*SZ,0.5*SZ,phi0+PI-semiSector,phi0+PI+semiSector,PIE);
    g.fill(0,255,0);
    g.arc(0.5*SZ,0.5*SZ,0.5*SZ,0.5*SZ,phi1-semiSector,phi1+semiSector,PIE);
    g.arc(0.5*SZ,0.5*SZ,0.5*SZ,0.5*SZ,phi1+PI-semiSector,phi1+PI+semiSector,PIE);
    g.fill(0,0,255);
    g.arc(0.5*SZ,0.5*SZ,0.5*SZ,0.5*SZ,phi2-semiSector,phi2+semiSector,PIE);
    g.arc(0.5*SZ,0.5*SZ,0.5*SZ,0.5*SZ,phi2+PI-semiSector,phi2+PI+semiSector,PIE);
    g.fill(255,127);
    g.ellipse(0.5*SZ,0.5*SZ,SZ/3.0,SZ/3.0);
    g.fill(255,196);
    g.ellipse(0.5*SZ,0.5*SZ,SZ/6.0,SZ/6.0);
  g.endDraw();
  return g;
}

// Shows a direction image
PImage directionImage(float[][] direction) {
  PImage im = new PImage(SZ,SZ);
  im.loadPixels();
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) 
    im.pixels[j*SZ+i] = directionColor(direction[i][j]);
  im.updatePixels();
  return im;
}

color directionColor(float direction) {
  direction = (1+direction/PI)*3.0; // From (-PI,+PI) to (0,6)
  int   i = floor(direction);
  float f = direction - i;
  color result = color(127);
  switch (i) {
    case 0: result = lerpColor(color(255, 255,   0),color(  0, 255,   0),f); break; // YELLOW  to GREEN
    case 1: result = lerpColor(color(0,   255,   0),color(  0, 255, 255),f); break; // GREEN   to CYAN
    case 2: result = lerpColor(color(0,   255, 255),color(  0,   0, 255),f); break; // CYAN    to BLUE
    case 3: result = lerpColor(color(0,     0, 255),color(255,   0, 255),f); break; // BLUE    to MAGENTA
    case 4: result = lerpColor(color(255,   0, 255),color(255,   0,   0),f); break; // MAGENTA to RED
    case 5: result = lerpColor(color(255,   0,   0),color(255, 255,   0),f); break; // RED     to YELLOW
    default: break;
  }
  return(result);
}

PImage directionColorScale() {
  PGraphics g = createGraphics(SZ,SZ,JAVA2D);
  g.beginDraw();
    g.background(0);
    g.ellipseMode(RADIUS);
    g.noStroke();
    for (int i=0; i<36; i++) {
      float phi1 = i*(TWO_PI/36.0)-PI;
      float phi2 = (i+1.0)*(TWO_PI/36.0)-PI;
      g.fill(directionColor(0.5*(phi1+phi2)));
      g.arc(0.5*SZ,0.5*SZ,0.5*SZ,0.5*SZ,phi1,phi2,PIE);
    }
    g.fill(255,127);
    g.ellipse(0.5*SZ,0.5*SZ,SZ/3.0,SZ/3.0);
    g.fill(255,196);
    g.ellipse(0.5*SZ,0.5*SZ,SZ/6.0,SZ/6.0);
  g.endDraw();
  return g;
}