static final float LEVEL = 0.25;

// Generate complex normal deviates
float[][] randomDataComplex() {
 float[][] data = new float[SZ][2*SZ];
 for (int i=0; i<SZ; i++) for (int j=0; j<2*SZ; j++)  data[i][j] = rnd.nextGaussian();
 return data;
}

// Normalize a data plane
float[][] zeroMeanUnitVariance(float[][] data) {
  float[] tmp = meanAndStandarddeviation(data);
  float mn = tmp[0];
  float sd = tmp[1];
  float[][] normalized = new float[SZ][SZ];
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++)  normalized[i][j] = (data[i][j]-mn)/sd;
  return normalized;
}

// This generates a random complex spectrum, multiplies it with a blur spectrum, and does an inverse FFT
// The real and imaginary parts are independent zero mean unit variance Gaussian noise fields 
ArrayList<float[][]> twoRandomFields(int theLevel) {
  // generate complex spectrum - Notice the NORMAL version is slower and you'll hardly notice a difference
  float[][] spectrum = randomDataComplex();
  // Remove DC-component
  spectrum[0][0] = 0.0;  
  // Multiply with kernel spectrum
  float[][] cs = complexMultiply(spectrum,blurSpectra.get(theLevel));
  // FFT back transform
  fft.complexInverse(cs,true);
  // Split real and imaginary parts
  float[][] re = new float[SZ][SZ];
  float[][] im = new float[SZ][SZ];
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) {
    re[i][j] = cs[i][2*j];
    im[i][j] = cs[i][2*j+1];
  }
  // return two independent noise fields
  ArrayList<float[][]> result = new ArrayList<float[][]>();
  result.add(zeroMeanUnitVariance(re));
  result.add(zeroMeanUnitVariance(im));
  return result;
}

// Random Gaussian noise stack, zero mean, unit variance
ArrayList<ArrayList<float[][]>> randomGaussianDataStacks() {
  ArrayList<ArrayList<float[][]>> noiseStacks = new ArrayList<ArrayList<float[][]>>();
  ArrayList<float[][]> noiseStackX = new ArrayList<float[][]>();
  ArrayList<float[][]> noiseStackY = new ArrayList<float[][]>();
  for (int k=0; k<NUM_SCALES; k++) {
    ArrayList<float[][]> tmp = twoRandomFields(k);
    noiseStackX.add(tmp.get(0));
    noiseStackY.add(tmp.get(1));
  }
  noiseStacks.add(noiseStackX);
  noiseStacks.add(noiseStackY);
  return noiseStacks;
}

// Random Gaussian noise stack, zero mean, standard deviation sigma
ArrayList<ArrayList<float[][]>> randomGaussianScaledDataStack() {
  ArrayList<ArrayList<float[][]>> incoherentNoiseStacks = randomGaussianDataStacks();
  for (int k=0; k<NUM_SCALES; k++) for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) {
    incoherentNoiseStacks.get(0).get(k)[i][j] *= effectiveScales[k];
    incoherentNoiseStacks.get(1).get(k)[i][j] *= effectiveScales[k];
  }
  return incoherentNoiseStacks;
}

// Coherent fractal noise stack, zero mean, unit variance
ArrayList<ArrayList<float[][]>> coherentGaussianDataStack() {
  ArrayList<ArrayList<float[][]>> incoherentNoiseStacks = randomGaussianDataStacks();
  ArrayList<float[][]> coherentNoiseStackX   = new ArrayList<float[][]>();
  ArrayList<float[][]> coherentNoiseStackY   = new ArrayList<float[][]>();
  // BUILD THE STACK IN REVERSE ORDER
  // Put the lowest resolution in the stack
  float[][] tmpX = new float[SZ][SZ];
  float[][] tmpY = new float[SZ][SZ];
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) {
    tmpX[i][j] = incoherentNoiseStacks.get(0).get(NUM_SCALES-1)[i][j];
    tmpY[i][j] = incoherentNoiseStacks.get(0).get(NUM_SCALES-1)[i][j];
  }
  coherentNoiseStackX.add(tmpX);
  coherentNoiseStackY.add(tmpY);
  // For each finer layer put the sum of the coarse fractal and the noise layer on the stack
  for (int k=NUM_SCALES-1; k>=0; k--) {
    int n = coherentNoiseStackX.size() - 1;
    float[][] tmp2X = new float[SZ][SZ];
    float[][] tmp2Y = new float[SZ][SZ];
    for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) {
      tmp2X[i][j] = incoherentNoiseStacks.get(0).get(k)[i][j] + coherentNoiseStackX.get(n)[i][j];
      tmp2Y[i][j] = incoherentNoiseStacks.get(1).get(k)[i][j] + coherentNoiseStackY.get(n)[i][j];
    }
    coherentNoiseStackX.add(tmp2X);
    coherentNoiseStackY.add(tmp2Y);
  }
  // PUT STACK UPSIDE DOWN (==RIGHT AGAIN!)
  ArrayList<float[][]> coherentNoiseStack2X = new ArrayList<float[][]>();
  ArrayList<float[][]> coherentNoiseStack2Y = new ArrayList<float[][]>();
  for (int k=NUM_SCALES-1; k>=0; k--) {
    coherentNoiseStack2X.add(coherentNoiseStackX.get(k));
    coherentNoiseStack2Y.add(coherentNoiseStackY.get(k));
  }
  ArrayList<ArrayList<float[][]>> coherentNoiseStacks2 = new ArrayList<ArrayList<float[][]>>();
  coherentNoiseStacks2.add(coherentNoiseStack2X);
  coherentNoiseStacks2.add(coherentNoiseStack2Y);
  return coherentNoiseStacks2;
}

// Estimate the "amplitude" of a stack
float stacksAmplitude(ArrayList<ArrayList<float[][]>> stacks, float level){
  FloatList d = new FloatList();
  for (int k=0; k<NUM_SCALES; k++) 
    for (int i=0; i<SZ; i++) 
      for (int j=0; j<SZ; j++) 
        d.append(sqrt(sq(stacks.get(0).get(k)[i][j]) + sq(stacks.get(1).get(k)[i][j])));
  d.sort();
  // return the quantile as a measure of "stack amplitude"
  return d.get(round(level*SZ*SZ));
}

// Normalise X, Y stacks to unit amplitude
ArrayList<ArrayList<float[][]>> normaliseStacks(ArrayList<ArrayList<float[][]>> stacks, float level) {
  float a = stacksAmplitude(stacks,level);
  ArrayList<ArrayList<float[][]>> stacksN = new ArrayList<ArrayList<float[][]>>();
  ArrayList<float[][]> stackX = new ArrayList<float[][]>();
  ArrayList<float[][]> stackY = new ArrayList<float[][]>();
  for (int k=0; k<NUM_SCALES; k++) {
    float[][] x = new float[SZ][SZ];
    float[][] y = new float[SZ][SZ];
    for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) {
      x[i][j] = stacks.get(0).get(k)[i][j]/a;
      y[i][j] = stacks.get(1).get(k)[i][j]/a;
    }
    stackX.add(x);
    stackY.add(y);
  }
  stacksN.add(stackX);
  stacksN.add(stackY);
  return stacksN;
}

void computeRandomDisarrayScaleSpaces() {
  initializationReport = initializationReport+"\n"+"Computing random disarray stacks";
  redraw();
  int epoch = millis();
  computeRandomDisarrayStacks();
  computeRandomScaledDisarrayStacks();
  computeCoherentDisarrayStacks();
  initializationReport = initializationReport+"\n"+"Computed random disarray stacks in "+(millis()-epoch)+" ms";
  redraw();
}

void computeRandomDisarrayStacks() {
  ArrayList<ArrayList<float[][]>> tmp = randomGaussianDataStacks();
  ArrayList<ArrayList<float[][]>> tmp2 = normaliseStacks(tmp,LEVEL);
  randomGaussianDataStackX.clear();
  randomGaussianDataStackY.clear();
  randomGaussianDataStackX = tmp2.get(0);
  randomGaussianDataStackY = tmp2.get(1);
}

void computeRandomScaledDisarrayStacks() {
  ArrayList<ArrayList<float[][]>> tmp = randomGaussianScaledDataStack();
  ArrayList<ArrayList<float[][]>> tmp2 = normaliseStacks(tmp,LEVEL);
  randomGaussianScaledDataStackX.clear();
  randomGaussianScaledDataStackY.clear();
  randomGaussianScaledDataStackX = tmp2.get(0);
  randomGaussianScaledDataStackY = tmp2.get(1);
}

void computeCoherentDisarrayStacks() {
  ArrayList<ArrayList<float[][]>> tmp = coherentGaussianDataStack();
  ArrayList<ArrayList<float[][]>> tmp2 = normaliseStacks(tmp,LEVEL);
  coherentDataStackX.clear();
  coherentDataStackY.clear();
  coherentDataStackX = tmp2.get(0);
  coherentDataStackY = tmp2.get(1);
}