float[] scales          = new float[NUM_SCALES];
float[] effectiveScales = new float[NUM_SCALES];
ArrayList<float[][]> blurSpectra = new ArrayList<float[][]>(); 

PImage               fiducialImage;                                            
float[][]            fiducialData;                                             
float[][]            fiducialSpectrum;  

ArrayList<float[][]> fiducialScaleSpace          = new ArrayList<float[][]>();
ArrayList<float[][]> fiducialScaleLayers         = new ArrayList<float[][]>();
ArrayList<float[][]> fiducialFirstOrderStackX    = new ArrayList<float[][]>();
ArrayList<float[][]> fiducialFirstOrderStackY    = new ArrayList<float[][]>();
ArrayList<float[][]> fiducialSecondOrderStackP   = new ArrayList<float[][]>();
ArrayList<float[][]> fiducialSecondOrderStackQ   = new ArrayList<float[][]>(); 
ArrayList<float[][]> fiducialSecondOrderStackR   = new ArrayList<float[][]>();

ArrayList<float[][]> randomGaussianDataStackX        = new ArrayList<float[][]>();
ArrayList<float[][]> randomGaussianDataStackY        = new ArrayList<float[][]>();
ArrayList<float[][]> randomGaussianScaledDataStackX  = new ArrayList<float[][]>();
ArrayList<float[][]> randomGaussianScaledDataStackY  = new ArrayList<float[][]>();
ArrayList<float[][]> coherentDataStackX              = new ArrayList<float[][]>(); 
ArrayList<float[][]> coherentDataStackY              = new ArrayList<float[][]>();

String initializationReport = "Initialization stage\n\nStarting with scale space";

// Set up everything connected with scales and the fiducial image
void getImage(String path) {
  int epoch = millis();
  // Try the image for size
  fiducialImage = loadImage(path);
  fiducialImage.resize(SZ,SZ); // Force standard size
  // Define scales
  scales[0] = 1.0/sqrt(2.0);
  for (int k=1; k<NUM_SCALES; k++) scales[k] = scales[k-1]*sqrt(2.0);
  for (int k=0; k<NUM_SCALES-1; k++) effectiveScales[k] = sqrt(scales[k]*scales[k+1]);
  effectiveScales[NUM_SCALES-1] =  effectiveScales[NUM_SCALES-2]*sqrt(2.0);
  initializationReport = initializationReport+"\n"+"Effective scales run from "+nf(effectiveScales[0],3,2)+" to "+nf(effectiveScales[NUM_SCALES-1],3,0);
  redraw();
  // store the image as float buffer and compute its complex spectrum
  fiducialData     = imageToData(fiducialImage);
  fiducialSpectrum = imageSpectrum(fiducialData);
  // compute the image scale space
  fiducialScaleSpace.clear();
  for (int i=0; i<NUM_SCALES; i++) fiducialScaleSpace.add(runConvolution(fiducialSpectrum,kernelSpectrum(scales[i])));
  // compute the image scale layer increments
  fiducialScaleLayers.clear();
  for (int k=0; k<NUM_SCALES-1; k++) {
    float[][] diff = new float[SZ][SZ];
    for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) diff[i][j] = fiducialScaleSpace.get(k)[i][j]-fiducialScaleSpace.get(k+1)[i][j];
    fiducialScaleLayers.add(diff);
  }
  float[][] diff = new float[SZ][SZ];
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) diff[i][j] = fiducialScaleSpace.get(NUM_SCALES-1)[i][j];
  fiducialScaleLayers.add(diff);
  // Compute complex blur spectra - this allows fast blurring when needed (AT LEAST IF USED!)
  blurSpectra.clear();
  for (int i=0; i<NUM_SCALES; i++) blurSpectra.add(kernelSpectrum(effectiveScales[i]));
  initializationReport = initializationReport+"\n"+"Set up took "+(millis()-epoch)+" ms";
  redraw();
}

void computeFirstOrderStack() {
  int epoch = millis();
  fiducialFirstOrderStackX.clear();
  fiducialFirstOrderStackY.clear();
  for (int k=0; k<NUM_SCALES; k++) { // Here I compute a Cartesian tangent vector basis
    float[][] x, y;
    fiducialFirstOrderStackX.add(runConvolution(fiducialSpectrum, kernelSpectrum(1,0,effectiveScales[k])));
    fiducialFirstOrderStackY.add(runConvolution(fiducialSpectrum, kernelSpectrum(0,1,effectiveScales[k])));
  }
  initializationReport = initializationReport+"\n"+"Computed first order in "+(millis()-epoch)+" ms";
  redraw();
}

void computeCorticalRepresentation(String path) {
  int epoch = millis();
  getImage(path); 
  computeRandomDisarrayScaleSpaces();
  initializationReport = initializationReport+"\n"+"Computing 1st order stack";
  redraw();
  computeFirstOrderStack();
  initializationReport = initializationReport+"\n"+"Computing 2nd order stack";
  redraw();
  computeSecondOrderStack();
  initializationReport = initializationReport+"\n"+"Computed cortical representation up to 2nd order in "+(millis()-epoch)+" ms";
  redraw();
  imageHasBeenDefined = true;
  firsttime = true;
}

void computeSecondOrderStack() {
  int epoch = millis();
  // Find the Hessiam matrix
  ArrayList<float[][]> hxx = new ArrayList<float[][]>();
  ArrayList<float[][]> hxy = new ArrayList<float[][]>();
  ArrayList<float[][]> hyy = new ArrayList<float[][]>();
  for (int k=0; k<NUM_SCALES; k++) { // Here I compute the coefficients of the Hessian matrix
    float[][] x, y;
    hxx.add(runConvolution(fiducialSpectrum, kernelSpectrum(2,0,effectiveScales[k])));
    hxy.add(runConvolution(fiducialSpectrum, kernelSpectrum(1,1,effectiveScales[k])));
    hyy.add(runConvolution(fiducialSpectrum, kernelSpectrum(0,2,effectiveScales[k])));
  }
  // Express in terms of line finder basis, directions n Pi/3, n = 0, 1, 2
  float z = 0.5*sqrt(3);
  fiducialSecondOrderStackP.clear();
  fiducialSecondOrderStackQ.clear();
  fiducialSecondOrderStackR.clear();
  for (int k=0; k<NUM_SCALES; k++) {
    float[][] tmp0 = new float[SZ][SZ];
    float[][] tmp1 = new float[SZ][SZ];
    float[][] tmp2 = new float[SZ][SZ];
    for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) {
      float zz = 0.25*hxx.get(k)[i][j]+0.75*hyy.get(k)[i][j]; // Here I transform from the Hessian matrix representation
      tmp0[i][j] =      hxx.get(k)[i][j];                     // To a nice, isotropic basis of three line finders at 60º orientation differences
      tmp1[i][j] = zz-z*hxy.get(k)[i][j];                     // The coefficients are easy enough to find with a little algebra
      tmp2[i][j] = zz+z*hxy.get(k)[i][j];
    }
    fiducialSecondOrderStackP.add(tmp0);
    fiducialSecondOrderStackQ.add(tmp1);
    fiducialSecondOrderStackR.add(tmp2);
  }
  initializationReport = initializationReport+"\n"+"Computed second order in "+(millis()-epoch)+" ms";
  redraw();
}

float[][] firstOrderDirectionality(float phi, int level) {
  float[][] result = new float[SZ][SZ];
  float c = cos(radians(phi));
  float s = sin(radians(phi));
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) 
    result[i][j] =c*fiducialFirstOrderStackX.get(level)[i][j]+s*fiducialFirstOrderStackY.get(level)[i][j];
  return result;
}

float[][] secondOrderDirectionality(float phi, int level) {
  float[][] result = new float[SZ][SZ];
  float u = (1.0+2.0*cos(2.0*radians(phi)))/3.0;                 // This is easy enough to prove with a little algebra
  float v =  4.0*cos(PI/6.0-radians(phi))*sin(radians(phi))/3.0;
  float w = -4.0*cos(PI/6.0+radians(phi))*sin(radians(phi))/3.0;
  for (int i=0; i<SZ; i++) for (int j=0; j<SZ; j++) 
    result[i][j] = u*fiducialSecondOrderStackP.get(level)[i][j]+v*fiducialSecondOrderStackQ.get(level)[i][j]+w*fiducialSecondOrderStackR.get(level)[i][j];
  return result;
}